# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import importlib
import os
import os.path as osp
import shlex
import warnings
import weakref
from enum import Enum, auto
from io import BytesIO
from typing import Any, Callable, Dict, Iterable, Iterator, Optional, Tuple, Union

import numpy as np

try:
    # Introduced in 1.20
    from numpy.typing import DTypeLike
except ImportError:
    DTypeLike = Any


class _IMAGE_BACKENDS(Enum):
    cv2 = auto()
    PIL = auto()


_IMAGE_BACKEND = None
_image_loading_errors = (FileNotFoundError,)
try:
    importlib.import_module("cv2")
    _IMAGE_BACKEND = _IMAGE_BACKENDS.cv2
except ModuleNotFoundError:
    import PIL

    _IMAGE_BACKEND = _IMAGE_BACKENDS.PIL
    _image_loading_errors = (*_image_loading_errors, PIL.UnidentifiedImageError)

from datumaro.util.image_cache import ImageCache
from datumaro.util.os_util import find_files


def __getattr__(name: str):
    if name in {"Image", "ByteImage"}:
        warnings.warn(
            f"Using {name} from 'util.image' is deprecated, "
            "the class is moved to 'components.media'",
            DeprecationWarning,
            stacklevel=2,
        )

        import datumaro.components.media as media_module

        return getattr(media_module, name)
    raise AttributeError(f"module {__name__} has no attribute {name}")


def load_image(path: str, dtype: DTypeLike = np.float32):
    """
    Reads an image in the HWC Grayscale/BGR(A) float [0; 255] format.
    """

    if _IMAGE_BACKEND == _IMAGE_BACKENDS.cv2:
        # cv2.imread does not support paths that are not representable
        # in the locale encoding on Windows, so we read the image bytes
        # ourselves.

        with open(path, "rb") as f:
            image_bytes = f.read()

        return decode_image(image_bytes, dtype=dtype)
    elif _IMAGE_BACKEND == _IMAGE_BACKENDS.PIL:
        from PIL import Image

        image = Image.open(path)
        image = np.asarray(image, dtype=dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image[:, :, :3] = image[:, :, 2::-1]  # RGB to BGR
    else:
        raise NotImplementedError()

    assert len(image.shape) in {2, 3}
    if len(image.shape) == 3:
        assert image.shape[2] in {3, 4}
    return image


def save_image(
    path: str, image: np.ndarray, create_dir: bool = False, dtype: DTypeLike = np.uint8, **kwargs
) -> None:
    # NOTE: Check destination path for existence
    # OpenCV silently fails if target directory does not exist
    dst_dir = osp.dirname(path)
    if dst_dir:
        if create_dir:
            os.makedirs(dst_dir, exist_ok=True)
        elif not osp.isdir(dst_dir):
            raise FileNotFoundError("Directory does not exist: '%s'" % dst_dir)

    if not kwargs:
        kwargs = {}

    # NOTE: OpenCV documentation says "If the image format is not supported,
    # the image will be converted to 8-bit unsigned and saved that way".
    # Conversion from np.int32 to np.uint8 is not working properly
    backend = _IMAGE_BACKEND
    if dtype == np.int32:
        backend = _IMAGE_BACKENDS.PIL
    if backend == _IMAGE_BACKENDS.cv2:
        # cv2.imwrite does not support paths that are not representable
        # in the locale encoding on Windows, so we write the image bytes
        # ourselves.

        ext = osp.splitext(path)[1]
        image_bytes = encode_image(image, ext, dtype=dtype, **kwargs)

        with open(path, "wb") as f:
            f.write(image_bytes)
    elif backend == _IMAGE_BACKENDS.PIL:
        from PIL import Image

        params = {}
        params["quality"] = kwargs.get("jpeg_quality")
        if kwargs.get("jpeg_quality") == 100:
            params["subsampling"] = 0

        image = image.astype(dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image[:, :, :3] = image[:, :, 2::-1]  # BGR to RGB
        image = Image.fromarray(image)
        image.save(path, **params)
    else:
        raise NotImplementedError()


def encode_image(image: np.ndarray, ext: str, dtype: DTypeLike = np.uint8, **kwargs) -> bytes:
    if not kwargs:
        kwargs = {}

    if _IMAGE_BACKEND == _IMAGE_BACKENDS.cv2:
        import cv2

        params = []

        if not ext.startswith("."):
            ext = "." + ext

        if ext.upper() == ".JPG":
            params = [int(cv2.IMWRITE_JPEG_QUALITY), kwargs.get("jpeg_quality", 75)]

        image = image.astype(dtype)
        success, result = cv2.imencode(ext, image, params=params)
        if not success:
            raise Exception("Failed to encode image to '%s' format" % (ext))
        return result.tobytes()
    elif _IMAGE_BACKEND == _IMAGE_BACKENDS.PIL:
        from PIL import Image

        if ext.startswith("."):
            ext = ext[1:]

        params = {}
        params["quality"] = kwargs.get("jpeg_quality")
        if kwargs.get("jpeg_quality") == 100:
            params["subsampling"] = 0

        image = image.astype(dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image[:, :, :3] = image[:, :, 2::-1]  # BGR to RGB
        image = Image.fromarray(image)
        with BytesIO() as buffer:
            image.save(buffer, format=ext, **params)
            return buffer.getvalue()
    else:
        raise NotImplementedError()


def decode_image(image_bytes: bytes, dtype: DTypeLike = np.float32) -> np.ndarray:
    if _IMAGE_BACKEND == _IMAGE_BACKENDS.cv2:
        import cv2

        image = np.frombuffer(image_bytes, dtype=np.uint8)
        image = cv2.imdecode(image, cv2.IMREAD_UNCHANGED)
        image = image.astype(dtype)
    elif _IMAGE_BACKEND == _IMAGE_BACKENDS.PIL:
        from PIL import Image

        image = Image.open(BytesIO(image_bytes))
        image = np.asarray(image, dtype=dtype)
        if len(image.shape) == 3 and image.shape[2] in {3, 4}:
            image[:, :, :3] = image[:, :, 2::-1]  # RGB to BGR
    else:
        raise NotImplementedError()

    assert len(image.shape) in {2, 3}
    if len(image.shape) == 3:
        assert image.shape[2] in {3, 4}
    return image


IMAGE_EXTENSIONS = {
    ".jpg",
    ".jpeg",
    ".jpe",
    ".jp2",
    ".png",
    ".bmp",
    ".dib",
    ".tif",
    ".tiff",
    ".tga",
    ".webp",
    ".pfm",
    ".sr",
    ".ras",
    ".exr",
    ".hdr",
    ".pic",
    ".pbm",
    ".pgm",
    ".ppm",
    ".pxm",
    ".pnm",
}


def find_images(
    dirpath: str,
    exts: Union[str, Iterable[str]] = None,
    recursive: bool = False,
    max_depth: int = None,
) -> Iterator[str]:
    yield from find_files(
        dirpath, exts=exts or IMAGE_EXTENSIONS, recursive=recursive, max_depth=max_depth
    )


def is_image(path: str) -> bool:
    trunk, ext = osp.splitext(osp.basename(path))
    return trunk and ext.lower() in IMAGE_EXTENSIONS and osp.isfile(path)


class lazy_image:
    def __init__(
        self,
        path: str,
        loader: Callable[[str], np.ndarray] = None,
        cache: Union[bool, ImageCache] = True,
    ) -> None:
        """
        Cache:
            - False: do not cache
            - True: use the global cache
            - ImageCache instance: an object to be used as cache
        """

        if loader is None:
            loader = load_image
        self._path = path
        self._loader = loader

        assert isinstance(cache, (ImageCache, bool))
        self._cache = cache

    def __call__(self) -> np.ndarray:
        image = None
        cache_key = weakref.ref(self)

        cache = self._get_cache()
        if cache is not None:
            image = cache.get(cache_key)

        if image is None:
            image = self._loader(self._path)
            if cache is not None:
                cache.push(cache_key, image)
        return image

    def _get_cache(self) -> Optional[ImageCache]:
        if self._cache is True:
            cache = ImageCache.get_instance()
        elif self._cache is False:
            cache = None
        else:
            cache = self._cache
        return cache


ImageMeta = Dict[str, Tuple[int, int]]
"""filename -> height, width"""

DEFAULT_IMAGE_META_FILE_NAME = "images.meta"


def load_image_meta_file(image_meta_path: str) -> ImageMeta:
    """
    Loads image metadata from a file with the following format:

        <image name 1> <height 1> <width 1>

        <image name 2> <height 2> <width 2>

        ...

    Shell-like comments and quoted fields are allowed.

    This can be useful to support datasets in which image dimensions are
    required to interpret annotations.
    """
    assert isinstance(image_meta_path, str)

    if not osp.isfile(image_meta_path):
        raise FileNotFoundError("Can't read image meta file '%s'" % image_meta_path)

    image_meta = {}

    with open(image_meta_path, encoding="utf-8") as f:
        for line in f:
            fields = shlex.split(line, comments=True)
            if not fields:
                continue

            # ignore extra fields, so that the format can be extended later
            image_name, h, w = fields[:3]
            image_meta[image_name] = (int(h), int(w))

    return image_meta


def save_image_meta_file(image_meta: ImageMeta, image_meta_path: str) -> None:
    """
    Saves image_meta to the path specified by image_meta_path in the format
    defined in load_image_meta_file's documentation.
    """

    assert isinstance(image_meta_path, str)

    with open(image_meta_path, "w", encoding="utf-8") as f:
        # Add a comment about file syntax
        print("# <image name> <height> <width>", file=f)
        print("", file=f)

        for image_name, (height, width) in image_meta.items():
            print(shlex.quote(image_name), height, width, file=f)



# Copyright (C) 2019-2020 Intel Corporation
#
# SPDX-License-Identifier: MIT


class CvatPath:
    IMAGES_DIR = "images"

    IMAGE_EXT = ".jpg"

    BUILTIN_ATTRS = {"occluded", "outside", "keyframe", "track_id"}



# Copyright (C) 2020-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import os
import os.path as osp
from collections import OrderedDict, defaultdict
from enum import Enum, auto
from itertools import chain
from typing import Dict, Optional, Set

from attrs import define, field

# Disable B410: import_lxml - the library is used for writing
from lxml import etree as ET  # nosec

from datumaro.components.annotation import (
    AnnotationType,
    Bbox,
    CompiledMask,
    Label,
    LabelCategories,
    Mask,
)
from datumaro.components.dataset import ItemStatus
from datumaro.components.dataset_base import DatasetItem
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.media import Image
from datumaro.util import find, str_to_bool
from datumaro.util.annotation_util import make_label_id_mapping
from datumaro.util.image import save_image
from datumaro.util.mask_tools import paint_mask, remap_mask
from datumaro.util.meta_file_util import has_meta_file

from .format import (
    VocInstColormap,
    VocPath,
    VocTask,
    make_voc_categories,
    make_voc_label_map,
    parse_label_map,
    parse_meta_file,
    write_label_map,
    write_meta_file,
)


def _convert_attr(name, attributes, type_conv, default=None):
    d = object()
    value = attributes.get(name, d)
    if value is d:
        return default

    try:
        return type_conv(value)
    except Exception as e:
        log.warning("Failed to convert attribute '%s'='%s': %s" % (name, value, e))
        return default


def _write_xml_bbox(bbox, parent_elem):
    x, y, w, h = bbox
    bbox_elem = ET.SubElement(parent_elem, "bndbox")
    ET.SubElement(bbox_elem, "xmin").text = str(x)
    ET.SubElement(bbox_elem, "ymin").text = str(y)
    ET.SubElement(bbox_elem, "xmax").text = str(x + w)
    ET.SubElement(bbox_elem, "ymax").text = str(y + h)
    return bbox_elem


class LabelmapType(Enum):
    voc = auto()
    source = auto()


@define
class _SubsetLists:
    class_lists: Dict[str, Optional[Set[int]]] = field(factory=dict)
    clsdet_list: Dict[str, Optional[bool]] = field(factory=dict)
    action_list: Dict[str, Optional[Dict[str, int]]] = field(factory=dict)
    layout_list: Dict[str, Optional[int]] = field(factory=dict)
    segm_list: Dict[str, Optional[bool]] = field(factory=dict)


class VocExporter(Exporter):
    DEFAULT_IMAGE_EXT = VocPath.IMAGE_EXT
    BUILTIN_ATTRS = {"difficult", "pose", "truncated", "occluded"}

    @staticmethod
    def _split_tasks_string(s):
        return [VocTask[i.strip()] for i in s.split(",")]

    @staticmethod
    def _get_labelmap(s):
        if osp.isfile(s):
            return s
        try:
            return LabelmapType[s].name
        except KeyError:
            import argparse

            raise argparse.ArgumentTypeError()

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)

        parser.add_argument(
            "--apply-colormap",
            type=str_to_bool,
            default=True,
            help="Use colormap for class and instance masks " "(default: %(default)s)",
        )
        parser.add_argument(
            "--label-map",
            type=cls._get_labelmap,
            default=None,
            help="Labelmap file path or one of %s" % ", ".join(t.name for t in LabelmapType),
        )
        parser.add_argument(
            "--allow-attributes",
            type=str_to_bool,
            default=True,
            help="Allow export of attributes (default: %(default)s)",
        )
        parser.add_argument(
            "--keep-empty",
            type=str_to_bool,
            default=False,
            help="Write subset lists even if they are empty " "(default: %(default)s)",
        )
        parser.add_argument(
            "--tasks",
            type=cls._split_tasks_string,
            help="VOC task filter, comma-separated list of {%s} "
            "(default: all)" % ", ".join(t.name for t in VocTask),
        )

        return parser

    def __init__(
        self,
        extractor,
        save_dir,
        tasks=None,
        apply_colormap=True,
        label_map=None,
        allow_attributes=True,
        keep_empty=False,
        **kwargs,
    ):
        super().__init__(extractor, save_dir, **kwargs)

        assert tasks is None or isinstance(tasks, (VocTask, list, set))
        if tasks is None:
            tasks = set(VocTask)
        elif isinstance(tasks, VocTask):
            tasks = {tasks}
        else:
            tasks = set(t if t in VocTask else VocTask[t] for t in tasks)
        self._tasks = tasks

        self._apply_colormap = apply_colormap
        self._allow_attributes = allow_attributes
        self._keep_empty = keep_empty

        if label_map is None:
            label_map = LabelmapType.source.name
        assert isinstance(label_map, (str, dict)), label_map
        self._load_categories(label_map)

        self._patch = None

    def apply(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        self.make_dirs()
        self.save_subsets()
        self.save_label_map()

    def make_dirs(self):
        save_dir = self._save_dir
        subsets_dir = osp.join(save_dir, VocPath.SUBSETS_DIR)
        cls_subsets_dir = osp.join(subsets_dir, VocPath.TASK_DIR[VocTask.classification])
        action_subsets_dir = osp.join(subsets_dir, VocPath.TASK_DIR[VocTask.action_classification])
        layout_subsets_dir = osp.join(subsets_dir, VocPath.TASK_DIR[VocTask.person_layout])
        segm_subsets_dir = osp.join(subsets_dir, VocPath.TASK_DIR[VocTask.segmentation])
        ann_dir = osp.join(save_dir, VocPath.ANNOTATIONS_DIR)
        img_dir = osp.join(save_dir, VocPath.IMAGES_DIR)
        segm_dir = osp.join(save_dir, VocPath.SEGMENTATION_DIR)
        inst_dir = osp.join(save_dir, VocPath.INSTANCES_DIR)
        images_dir = osp.join(save_dir, VocPath.IMAGES_DIR)

        os.makedirs(subsets_dir, exist_ok=True)
        os.makedirs(ann_dir, exist_ok=True)
        os.makedirs(img_dir, exist_ok=True)
        os.makedirs(segm_dir, exist_ok=True)
        os.makedirs(inst_dir, exist_ok=True)
        os.makedirs(images_dir, exist_ok=True)

        self._subsets_dir = subsets_dir
        self._cls_subsets_dir = cls_subsets_dir
        self._action_subsets_dir = action_subsets_dir
        self._layout_subsets_dir = layout_subsets_dir
        self._segm_subsets_dir = segm_subsets_dir
        self._ann_dir = ann_dir
        self._img_dir = img_dir
        self._segm_dir = segm_dir
        self._inst_dir = inst_dir
        self._images_dir = images_dir

    def get_label(self, label_id):
        return self._extractor.categories()[AnnotationType.label].items[label_id].name

    def save_subsets(self):
        subsets = self._extractor.subsets()
        pbars = self._ctx.progress_reporter.split(len(subsets))
        for pbar, (subset_name, subset) in zip(pbars, subsets.items()):
            lists = _SubsetLists()

            for item in pbar.iter(subset, desc=f"Exporting '{subset_name}'"):
                log.debug("Converting item '%s'", item.id)

                try:
                    image_filename = self._make_image_filename(item)
                    if self._save_media:
                        if item.media:
                            self._save_image(item, osp.join(self._images_dir, image_filename))
                        else:
                            log.debug("Item '%s' has no image", item.id)

                    self._export_annotations(item, image_filename=image_filename, lists=lists)
                except Exception as e:
                    self._ctx.error_policy.report_item_error(e, item_id=(item.id, item.subset))

            if self._tasks & {
                VocTask.classification,
                VocTask.detection,
                VocTask.action_classification,
                VocTask.person_layout,
            }:
                self.save_clsdet_lists(subset_name, lists.clsdet_list)
                if self._tasks & {VocTask.classification}:
                    self.save_class_lists(subset_name, lists.class_lists)
            if self._tasks & {VocTask.action_classification}:
                self.save_action_lists(subset_name, lists.action_list)
            if self._tasks & {VocTask.person_layout}:
                self.save_layout_lists(subset_name, lists.layout_list)
            if self._tasks & {VocTask.segmentation}:
                self.save_segm_lists(subset_name, lists.segm_list)

    def _export_annotations(self, item: DatasetItem, *, image_filename: str, lists: _SubsetLists):
        labels = []
        bboxes = []
        masks = []
        for a in item.annotations:
            if isinstance(a, Label):
                labels.append(a)
            elif isinstance(a, Bbox):
                bboxes.append(a)
            elif isinstance(a, Mask):
                masks.append(a)

        if self._tasks & {VocTask.detection, VocTask.person_layout, VocTask.action_classification}:
            root_elem = ET.Element("annotation")
            if "_" in item.id:
                folder = item.id[: item.id.find("_")]
            else:
                folder = ""
            ET.SubElement(root_elem, "folder").text = folder
            ET.SubElement(root_elem, "filename").text = image_filename

            source_elem = ET.SubElement(root_elem, "source")
            ET.SubElement(source_elem, "database").text = "Unknown"
            ET.SubElement(source_elem, "annotation").text = "Unknown"
            ET.SubElement(source_elem, "image").text = "Unknown"

            if item.media and item.media.has_size:
                h, w = item.media.size
                size_elem = ET.SubElement(root_elem, "size")
                ET.SubElement(size_elem, "width").text = str(w)
                ET.SubElement(size_elem, "height").text = str(h)
                depth = ""
                if item.media.data is not None:
                    depth = str(item.media.data.shape[-1])
                ET.SubElement(size_elem, "depth").text = depth

            item_segmented = 0 < len(masks)
            ET.SubElement(root_elem, "segmented").text = str(int(item_segmented))

            objects_with_parts = []
            objects_with_actions = defaultdict(dict)

            main_bboxes = []
            layout_bboxes = []
            for bbox in bboxes:
                label = self.get_label(bbox.label)
                if self._is_part(label):
                    layout_bboxes.append(bbox)
                elif self._is_label(label):
                    main_bboxes.append(bbox)

            for new_obj_id, obj in enumerate(main_bboxes):
                attr = obj.attributes

                obj_elem = ET.SubElement(root_elem, "object")

                obj_label = self.get_label(obj.label)
                ET.SubElement(obj_elem, "name").text = obj_label

                if "pose" in attr:
                    ET.SubElement(obj_elem, "pose").text = str(attr["pose"])

                ET.SubElement(obj_elem, "truncated").text = "%d" % _convert_attr(
                    "truncated", attr, int, 0
                )
                ET.SubElement(obj_elem, "occluded").text = "%d" % _convert_attr(
                    "occluded", attr, int, 0
                )
                ET.SubElement(obj_elem, "difficult").text = "%d" % _convert_attr(
                    "difficult", attr, int, 0
                )

                bbox = obj.get_bbox()
                if bbox is not None:
                    _write_xml_bbox(bbox, obj_elem)

                for part_bbox in filter(
                    lambda x: obj.group and obj.group == x.group, layout_bboxes
                ):
                    part_elem = ET.SubElement(obj_elem, "part")
                    ET.SubElement(part_elem, "name").text = self.get_label(part_bbox.label)
                    _write_xml_bbox(part_bbox.get_bbox(), part_elem)

                    objects_with_parts.append(new_obj_id)

                label_actions = self._get_actions(obj_label)
                actions_elem = ET.Element("actions")
                for action in label_actions:
                    present = 0
                    if action in attr:
                        present = _convert_attr(action, attr, lambda v: int(v is True), 0)
                        if action.isdigit():
                            action = "_" + action
                        action = action.replace(" ", "_")
                        ET.SubElement(actions_elem, action).text = "%d" % present

                    objects_with_actions[new_obj_id][action] = present
                if len(actions_elem) != 0:
                    obj_elem.append(actions_elem)

                if self._allow_attributes:
                    native_attrs = set(self.BUILTIN_ATTRS)
                    native_attrs.update(label_actions)

                    attrs_elem = ET.Element("attributes")
                    for k, v in attr.items():
                        if k in native_attrs:
                            continue
                        attr_elem = ET.SubElement(attrs_elem, "attribute")
                        ET.SubElement(attr_elem, "name").text = str(k)
                        ET.SubElement(attr_elem, "value").text = str(v)
                    if len(attrs_elem):
                        obj_elem.append(attrs_elem)

            if self._tasks & {
                VocTask.detection,
                VocTask.person_layout,
                VocTask.action_classification,
            }:
                ann_path = osp.join(self._ann_dir, item.id + ".xml")
                os.makedirs(osp.dirname(ann_path), exist_ok=True)
                with open(ann_path, "w", encoding="utf-8") as f:
                    f.write(ET.tostring(root_elem, encoding="unicode", pretty_print=True))

            lists.clsdet_list[item.id] = True

            if objects_with_parts:
                lists.layout_list[item.id] = objects_with_parts

            if objects_with_actions:
                lists.action_list[item.id] = objects_with_actions

        for label_ann in labels:
            label = self.get_label(label_ann.label)
            if not self._is_label(label):
                continue
            class_list = lists.class_lists.get(item.id, set())
            class_list.add(label_ann.label)
            lists.class_lists[item.id] = class_list

            lists.clsdet_list[item.id] = True

        if masks and VocTask.segmentation in self._tasks:
            compiled_mask = CompiledMask.from_instance_masks(
                masks, instance_labels=[self._label_id_mapping(m.label) for m in masks]
            )

            self.save_segm(
                osp.join(self._segm_dir, item.id + VocPath.SEGM_EXT), compiled_mask.class_mask
            )
            self.save_segm(
                osp.join(self._inst_dir, item.id + VocPath.SEGM_EXT),
                compiled_mask.instance_mask,
                colormap=VocInstColormap,
            )

            lists.segm_list[item.id] = True
        elif not masks and self._patch:
            cls_mask_path = osp.join(self._segm_dir, item.id + VocPath.SEGM_EXT)
            if osp.isfile(cls_mask_path):
                os.remove(cls_mask_path)

            inst_mask_path = osp.join(self._inst_dir, item.id + VocPath.SEGM_EXT)
            if osp.isfile(inst_mask_path):
                os.remove(inst_mask_path)

        if len(item.annotations) == 0:
            lists.clsdet_list[item.id] = None
            lists.layout_list[item.id] = None
            lists.action_list[item.id] = None
            lists.segm_list[item.id] = None

    @staticmethod
    def _get_filtered_lines(path, patch, subset, items=None):
        lines = {}
        with open(path, encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if not line:
                    continue
                line_parts = line.split(maxsplit=1)
                if len(line_parts) < 2:
                    line_parts.append("")
                item, text = line_parts
                if not patch or patch.updated_items.get((item, subset)) != ItemStatus.removed:
                    lines.setdefault(item, []).append(text)
        if items is not None:
            items.update((k, True) for k in lines)
        return lines

    def save_action_lists(self, subset_name, action_list):
        os.makedirs(self._action_subsets_dir, exist_ok=True)

        ann_file = osp.join(self._action_subsets_dir, subset_name + ".txt")
        items = {k: True for k in action_list}
        if self._patch and osp.isfile(ann_file):
            self._get_filtered_lines(ann_file, self._patch, subset_name, items)

        if items or self._keep_empty:
            with open(ann_file, "w", encoding="utf-8") as f:
                for item in items:
                    f.write("%s\n" % item)
        elif osp.isfile(ann_file):
            os.remove(ann_file)

        if not items and not self._patch and not self._keep_empty:
            return

        def _write_item(f, item, objs, action):
            if not objs:
                return
            for obj_id, obj_actions in objs.items():
                presented = obj_actions.get(action)
                f.write("%s %s % d\n" % (item, 1 + obj_id, 1 if presented else -1))

        all_actions = {
            act: osp.join(self._action_subsets_dir, "%s_%s.txt" % (act, subset_name))
            for act in chain(*(self._get_actions(l) for l in self._label_map))
        }
        for action, ann_file in all_actions.items():
            if not items and not self._keep_empty:
                if osp.isfile(ann_file):
                    os.remove(ann_file)
                continue

            lines = {}
            if self._patch and osp.isfile(ann_file):
                lines = self._get_filtered_lines(ann_file, None, subset_name)

            with open(ann_file, "w", encoding="utf-8") as f:
                for item in items:
                    if item in action_list:
                        _write_item(f, item, action_list[item], action)
                    elif item in lines:
                        print(item, *lines[item], file=f)

    def save_class_lists(self, subset_name, class_lists):
        def _write_item(f, item, item_labels):
            if not item_labels:
                return
            item_labels = [self.get_label(l) for l in item_labels]
            presented = label in item_labels
            f.write("%s % d\n" % (item, 1 if presented else -1))

        os.makedirs(self._cls_subsets_dir, exist_ok=True)

        for label in self._label_map:
            ann_file = osp.join(self._cls_subsets_dir, "%s_%s.txt" % (label, subset_name))
            items = {k: True for k in class_lists}
            lines = {}
            if self._patch and osp.isfile(ann_file):
                lines = self._get_filtered_lines(ann_file, self._patch, subset_name, items)

            if not items and not self._keep_empty:
                if osp.isfile(ann_file):
                    os.remove(ann_file)
                continue

            with open(ann_file, "w", encoding="utf-8") as f:
                for item in items:
                    if item in class_lists:
                        _write_item(f, item, class_lists[item])
                    elif item in lines:
                        print(item, *lines[item], file=f)

    def save_clsdet_lists(self, subset_name, clsdet_list):
        os.makedirs(self._cls_subsets_dir, exist_ok=True)

        ann_file = osp.join(self._cls_subsets_dir, subset_name + ".txt")
        items = {k: True for k in clsdet_list}
        if self._patch and osp.isfile(ann_file):
            self._get_filtered_lines(ann_file, self._patch, subset_name, items)

        if items or self._keep_empty:
            with open(ann_file, "w", encoding="utf-8") as f:
                for item in items:
                    f.write("%s\n" % item)
        elif osp.isfile(ann_file):
            os.remove(ann_file)

    def save_segm_lists(self, subset_name, segm_list):
        os.makedirs(self._segm_subsets_dir, exist_ok=True)

        ann_file = osp.join(self._segm_subsets_dir, subset_name + ".txt")
        items = {k: True for k in segm_list}
        if self._patch and osp.isfile(ann_file):
            self._get_filtered_lines(ann_file, self._patch, subset_name, items)

        if items or self._keep_empty:
            with open(ann_file, "w", encoding="utf-8") as f:
                for item in items:
                    f.write("%s\n" % item)
        elif osp.isfile(ann_file):
            os.remove(ann_file)

    def save_layout_lists(self, subset_name, layout_list):
        def _write_item(f, item, item_layouts):
            if 1 < len(item.split()):
                item = '"' + item + '"'
            if item_layouts:
                for obj_id in item_layouts:
                    f.write("%s % d\n" % (item, 1 + obj_id))
            else:
                f.write("%s\n" % item)

        os.makedirs(self._layout_subsets_dir, exist_ok=True)

        ann_file = osp.join(self._layout_subsets_dir, subset_name + ".txt")
        items = {k: True for k in layout_list}
        lines = {}
        if self._patch and osp.isfile(ann_file):
            self._get_filtered_lines(ann_file, self._patch, subset_name, items)

        if not items and not self._keep_empty:
            if osp.isfile(ann_file):
                os.remove(ann_file)
            return

        with open(ann_file, "w", encoding="utf-8") as f:
            for item in items:
                if item in layout_list:
                    _write_item(f, item, layout_list[item])
                elif item in lines:
                    print(item, *lines[item], file=f)

    def save_segm(self, path, mask, colormap=None):
        if self._apply_colormap:
            if colormap is None:
                colormap = self._categories[AnnotationType.mask].colormap
            mask = paint_mask(mask, colormap)
        save_image(path, mask, create_dir=True)

    def save_label_map(self):
        if self._save_dataset_meta:
            write_meta_file(self._save_dir, self._label_map)
        else:
            path = osp.join(self._save_dir, VocPath.LABELMAP_FILE)
            write_label_map(path, self._label_map)

    def _load_categories(self, label_map_source):
        if label_map_source == LabelmapType.voc.name:
            # use the default VOC colormap
            label_map = make_voc_label_map()

        elif (
            label_map_source == LabelmapType.source.name
            and AnnotationType.mask not in self._extractor.categories()
        ):
            # generate colormap for input labels
            labels = self._extractor.categories().get(AnnotationType.label, LabelCategories())
            label_map = OrderedDict((item.name, [None, [], []]) for item in labels.items)

        elif (
            label_map_source == LabelmapType.source.name
            and AnnotationType.mask in self._extractor.categories()
        ):
            # use source colormap
            labels = self._extractor.categories()[AnnotationType.label]
            colors = self._extractor.categories()[AnnotationType.mask]
            label_map = OrderedDict()
            for idx, item in enumerate(labels.items):
                color = colors.colormap.get(idx)
                if color is not None:
                    label_map[item.name] = [color, [], []]

        elif isinstance(label_map_source, dict):
            label_map = OrderedDict(sorted(label_map_source.items(), key=lambda e: e[0]))

        elif isinstance(label_map_source, str) and osp.isfile(label_map_source):
            if has_meta_file(label_map_source):
                label_map = parse_meta_file(label_map_source)
            else:
                label_map = parse_label_map(label_map_source)

        else:
            raise Exception(
                "Wrong labelmap specified: '%s', "
                "expected one of %s or a file path"
                % (label_map_source, ", ".join(t.name for t in LabelmapType))
            )

        bg_label = find(label_map.items(), lambda x: x[1][0] == (0, 0, 0))
        if bg_label is None:
            bg_label = "background"
            if bg_label not in label_map:
                has_colors = any(v[0] is not None for v in label_map.values())
                color = (0, 0, 0) if has_colors else None
                label_map[bg_label] = [color, [], []]
            label_map.move_to_end(bg_label, last=False)

        self._categories = make_voc_categories(label_map)

        # Update colors with assigned values
        colormap = self._categories[AnnotationType.mask].colormap
        for label_id, color in colormap.items():
            label_desc = label_map[self._categories[AnnotationType.label].items[label_id].name]
            label_desc[0] = color

        self._label_map = label_map
        self._label_id_mapping = self._make_label_id_map()

    def _is_label(self, s):
        return self._label_map.get(s) is not None

    def _is_part(self, s):
        for label_desc in self._label_map.values():
            if s in label_desc[1]:
                return True
        return False

    def _is_action(self, label, s):
        return s in self._get_actions(label)

    def _get_actions(self, label):
        label_desc = self._label_map.get(label)
        if not label_desc:
            return []
        return label_desc[2]

    def _make_label_id_map(self):
        map_id, id_mapping, src_labels, dst_labels = make_label_id_mapping(
            self._extractor.categories().get(AnnotationType.label),
            self._categories[AnnotationType.label],
        )

        void_labels = [
            src_label for src_id, src_label in src_labels.items() if src_label not in dst_labels
        ]
        if void_labels:
            log.warning(
                "The following labels are remapped to background: %s" % ", ".join(void_labels)
            )
        log.debug(
            "Saving segmentations with the following label mapping: \n%s"
            % "\n".join(
                [
                    "#%s '%s' -> #%s '%s'"
                    % (
                        src_id,
                        src_label,
                        id_mapping[src_id],
                        self._categories[AnnotationType.label].items[id_mapping[src_id]].name,
                    )
                    for src_id, src_label in src_labels.items()
                ]
            )
        )

        return map_id

    def _remap_mask(self, mask):
        return remap_mask(mask, self._label_id_mapping)

    @classmethod
    def patch(cls, dataset, patch, save_dir, **kwargs):
        conv = cls(patch.as_dataset(dataset), save_dir=save_dir, **kwargs)
        conv._patch = patch
        conv.apply()

        for filename in os.listdir(conv._cls_subsets_dir):
            if "_" not in filename or not filename.endswith(".txt"):
                continue

            label, subset = osp.splitext(filename)[0].split("_", maxsplit=1)
            if label not in conv._label_map or subset not in dataset.subsets():
                os.remove(osp.join(conv._cls_subsets_dir, filename))

        # Find images that need to be removed
        # images from different subsets are stored in the common directory
        # Avoid situations like:
        # (a, test): added
        # (a, train): removed
        # where the second line removes images from the first.
        ids_to_remove = {}
        for (item_id, subset), status in patch.updated_items.items():
            if status != ItemStatus.removed:
                item = patch.data.get(item_id, subset)
            else:
                item = DatasetItem(item_id, subset=subset)

            if not (status == ItemStatus.removed or not item.media):
                ids_to_remove[item_id] = (item, False)
            else:
                ids_to_remove.setdefault(item_id, (item, True))

        for item, to_remove in ids_to_remove.values():
            if not to_remove:
                continue

            if conv._tasks & {
                VocTask.detection,
                VocTask.action_classification,
                VocTask.person_layout,
            }:
                ann_path = osp.join(conv._ann_dir, item.id + ".xml")
                if osp.isfile(ann_path):
                    os.remove(ann_path)

            image_path = osp.join(conv._images_dir, conv._make_image_filename(item))
            if osp.isfile(image_path):
                os.unlink(image_path)

            if not [a for a in item.annotations if a.type is AnnotationType.mask]:
                path = osp.join(save_dir, VocPath.SEGMENTATION_DIR, item.id + VocPath.SEGM_EXT)
                if osp.isfile(path):
                    os.unlink(path)

                path = osp.join(save_dir, VocPath.INSTANCES_DIR, item.id + VocPath.SEGM_EXT)
                if osp.isfile(path):
                    os.unlink(path)


class VocClassificationExporter(VocExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = VocTask.classification
        super().__init__(*args, **kwargs)


class VocDetectionExporter(VocExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = VocTask.detection
        super().__init__(*args, **kwargs)


class VocLayoutExporter(VocExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = VocTask.person_layout
        super().__init__(*args, **kwargs)


class VocActionExporter(VocExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = VocTask.action_classification
        super().__init__(*args, **kwargs)


class VocSegmentationExporter(VocExporter):
    def __init__(self, *args, **kwargs):
        kwargs["tasks"] = VocTask.segmentation
        super().__init__(*args, **kwargs)



# Copyright (C) 2019-2020 Intel Corporation
#
# SPDX-License-Identifier: MIT

import numpy as np

from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.transformer import Transform
from datumaro.util import take_by


# pylint: disable=no-self-use
class Launcher(CliPlugin):
    def __init__(self, model_dir=None):
        pass

    def launch(self, inputs):
        raise NotImplementedError()

    def infos(self):
        return None

    def categories(self):
        return None


# pylint: enable=no-self-use


class ModelTransform(Transform):
    def __init__(self, extractor, launcher, batch_size=1):
        super().__init__(extractor)
        self._launcher = launcher
        self._batch_size = batch_size

    def __iter__(self):
        for batch in take_by(self._extractor, self._batch_size):
            inputs = np.array([np.atleast_3d(item.media.data) for item in batch])
            inference = self._launcher.launch(inputs)

            for item, annotations in zip(batch, inference):
                self._check_annotations(annotations)
                yield self.wrap_item(item, annotations=annotations)

    def get_subset(self, name):
        subset = self._extractor.get_subset(name)
        return __class__(subset, self._launcher, self._batch_size)

    def infos(self):
        launcher_override = self._launcher.infos()
        if launcher_override is not None:
            return launcher_override
        return self._extractor.infos()

    def categories(self):
        launcher_override = self._launcher.categories()
        if launcher_override is not None:
            return launcher_override
        return self._extractor.categories()

    def transform_item(self, item):
        inputs = np.expand_dims(item.media, axis=0)
        annotations = self._launcher.launch(inputs)[0]
        return self.wrap_item(item, annotations=annotations)

    def _check_annotations(self, annotations):
        labels_count = len(self.categories().get(AnnotationType.label, LabelCategories()).items)

        for ann in annotations:
            label = getattr(ann, "label")
            if label is None:
                continue

            if label not in range(labels_count):
                raise Exception(
                    "Annotation has unexpected label id %s, "
                    "while there is only %s defined labels." % (label, labels_count)
                )



# Copyright (C) 2021-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.components.extractor_tfds import AVAILABLE_TFDS_DATASETS, TFDS_EXTRACTOR_AVAILABLE
from datumaro.components.project import Environment
from datumaro.util.os_util import make_file_name

from ..util import MultilineFormatter
from ..util.errors import CliException
from ..util.project import generate_next_file_name


def build_parser(parser_ctor=argparse.ArgumentParser):
    builtin_writers = sorted(Environment().exporters)
    if TFDS_EXTRACTOR_AVAILABLE:
        available_datasets = ", ".join(f"tfds:{name}" for name in AVAILABLE_TFDS_DATASETS)
    else:
        available_datasets = "N/A (TensorFlow and/or TensorFlow Datasets " "are not installed)"

    parser = parser_ctor(
        help="Download a publicly available dataset",
        description="""
        Downloads a publicly available dataset and saves it in a given format.|n
        |n
        Currently, the only source of datasets is the TensorFlow Datasets
        library. Therefore, to use this command you must install TensorFlow &
        TFDS, which you can do as follows:|n
        |n
        |s|spip install datumaro[tf,tfds]|n
        |n
        Supported datasets: {}|n
        |n
        For information about the datasets, run "datum describe-downloads".
        More detailed information can be found in the TFDS Catalog:
        <https://www.tensorflow.org/datasets/catalog/overview>.|n
        |n
        Supported output formats: {}|n
        |n
        Examples:|n
        - Download the MNIST dataset:|n
        |s|s%(prog)s -i tfds:mnist -- --save-media|n
        |n
        - Download the VOC 2012 dataset, saving only the annotations in the COCO
          format into a specific directory:|n
        |s|s%(prog)s -i tfds:voc/2012 -f coco -o path/I/like/
        """.format(
            available_datasets, ", ".join(builtin_writers)
        ),
        formatter_class=MultilineFormatter,
    )

    parser.add_argument("-i", "--dataset-id", required=True, help="Which dataset to download")
    parser.add_argument(
        "-f", "--output-format", help="Output format (default: original format of the dataset)"
    )
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="dst_dir",
        help="Directory to save output (default: a subdir in the current one)",
    )
    parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite existing files in the save directory"
    )
    parser.add_argument("-s", "--subset", help="Save only the specified subset")
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Additional arguments for output format (pass '-- -h' for help). "
        "Must be specified after the main command arguments",
    )
    parser.set_defaults(command=download_command)

    return parser


def get_sensitive_args():
    return {
        download_command: ["dst_dir", "extra_args"],
    }


def download_command(args):
    env = Environment()

    if args.dataset_id.startswith("tfds:"):
        if TFDS_EXTRACTOR_AVAILABLE:
            tfds_ds_name = args.dataset_id[5:]
            tfds_ds = AVAILABLE_TFDS_DATASETS.get(tfds_ds_name)
            if tfds_ds:
                default_output_format = tfds_ds.metadata.default_output_format
                extractor_factory = tfds_ds.make_extractor
            else:
                raise CliException(f"Unsupported TFDS dataset '{tfds_ds_name}'")
        else:
            raise CliException(
                "TFDS datasets are not available, because TFDS and/or "
                "TensorFlow are not installed.\n"
                "You can install them with: pip install datumaro[tf,tfds]"
            )
    else:
        raise CliException(f"Unknown dataset ID '{args.dataset_id}'")

    output_format = args.output_format or default_output_format

    try:
        exporter = env.exporters[output_format]
    except KeyError:
        raise CliException("Exporter for format '%s' is not found" % output_format)
    extra_args = exporter.parse_cmdline(args.extra_args)

    dst_dir = args.dst_dir
    if dst_dir:
        if not args.overwrite and osp.isdir(dst_dir) and os.listdir(dst_dir):
            raise CliException(
                "Directory '%s' already exists " "(pass --overwrite to overwrite)" % dst_dir
            )
    else:
        dst_dir = generate_next_file_name(
            "%s-%s"
            % (
                make_file_name(args.dataset_id),
                make_file_name(output_format),
            )
        )
    dst_dir = osp.abspath(dst_dir)

    log.info("Downloading the dataset")
    extractor = extractor_factory()

    if args.subset:
        try:
            extractor = extractor.subsets()[args.subset]
        except KeyError:
            raise CliException("Subset '%s' is not present in the dataset" % args.subset)

    log.info("Exporting the dataset")
    exporter.convert(extractor, dst_dir, default_image_ext=".png", **extra_args)

    log.info("Dataset exported to '%s' as '%s'" % (dst_dir, output_format))



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp
from enum import Enum


class KittiRawPath:
    PCD_DIR = osp.join("velodyne_points", "data")
    IMG_DIR_PREFIX = "image_"
    BUILTIN_ATTRS = {"occluded", "truncation", "occlusion"}
    SPECIAL_ATTRS = {
        "track_id",
    }
    ANNO_FILE = "tracklet_labels.xml"
    NAME_MAPPING_FILE = "frame_list.txt"


class PoseStates(Enum):
    UNSET = 0
    INTERP = 1
    LABELED = 2


class OcclusionStates(Enum):
    OCCLUSION_UNSET = -1
    VISIBLE = 0
    PARTLY = 1
    FULLY = 2


class TruncationStates(Enum):
    TRUNCATION_UNSET = -1
    IN_IMAGE = 0
    TRUNCATED = 1
    OUT_IMAGE = 2
    BEHIND_IMAGE = 99



# Copyright (C) 2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os.path as osp

import numpy as np

from datumaro.components.annotation import AnnotationType, Cuboid3d, LabelCategories, Mask
from datumaro.components.dataset_base import DatasetItem, SubsetBase
from datumaro.components.format_detection import FormatDetectionContext
from datumaro.components.importer import Importer
from datumaro.components.media import MultiframeImage
from datumaro.util.pickle_util import PickleLoader


class BratsNumpyPath:
    IDS_FILE = "val_ids.p"
    BOXES_FILE = "val_brain_bbox.p"
    LABELS_FILE = "labels"
    DATA_SUFFIX = "_data_cropped"
    LABEL_SUFFIX = "_label_cropped"


class BratsNumpyBase(SubsetBase):
    def __init__(self, path):
        if not osp.isfile(path):
            raise FileNotFoundError("Can't read annotation file '%s'" % path)

        super().__init__(media_type=MultiframeImage)

        self._root_dir = osp.dirname(path)
        self._categories = self._load_categories()
        self._items = list(self._load_items(path).values())

    def _load_categories(self):
        label_cat = LabelCategories()

        labels_path = osp.join(self._root_dir, BratsNumpyPath.LABELS_FILE)
        if osp.isfile(labels_path):
            with open(labels_path, encoding="utf-8") as f:
                for line in f:
                    label_cat.add(line.strip())

        return {AnnotationType.label: label_cat}

    def _load_items(self, path):
        items = {}

        with open(path, "rb") as f:
            ids = PickleLoader.restricted_load(f)

        boxes = None
        boxes_file = osp.join(self._root_dir, BratsNumpyPath.BOXES_FILE)
        if osp.isfile(boxes_file):
            with open(boxes_file, "rb") as f:
                boxes = PickleLoader.restricted_load(f)

        for i, item_id in enumerate(ids):
            image_path = osp.join(self._root_dir, item_id + BratsNumpyPath.DATA_SUFFIX + ".npy")
            media = None
            if osp.isfile(image_path):
                data = np.load(image_path)[0].transpose()
                images = [0] * data.shape[2]
                for j in range(data.shape[2]):
                    images[j] = data[:, :, j]

                media = MultiframeImage(images, path=image_path)

            anno = []
            mask_path = osp.join(self._root_dir, item_id + BratsNumpyPath.LABEL_SUFFIX + ".npy")
            if osp.isfile(mask_path):
                mask = np.load(mask_path)[0].transpose()
                for j in range(mask.shape[2]):
                    classes = np.unique(mask[:, :, j])
                    for class_id in classes:
                        anno.append(
                            Mask(
                                image=self._lazy_extract_mask(mask[:, :, j], class_id),
                                label=class_id,
                                attributes={"image_id": j},
                            )
                        )

            if boxes is not None:
                box = boxes[i]
                anno.append(Cuboid3d(position=list(box[0]), rotation=list(box[1])))

            items[item_id] = DatasetItem(id=item_id, media=media, annotations=anno)

        return items

    @staticmethod
    def _lazy_extract_mask(mask, c):
        return lambda: mask == c


class BratsNumpyImporter(Importer):
    @classmethod
    def detect(cls, context: FormatDetectionContext) -> None:
        context.require_file(BratsNumpyPath.IDS_FILE)

    @classmethod
    def find_sources(cls, path):
        return cls._find_sources_recursive(
            path, "", "brats_numpy", filename=BratsNumpyPath.IDS_FILE
        )



# Copyright (C) 2019-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import os
import os.path as osp
import shutil
import warnings
from tempfile import mkdtemp
from typing import NoReturn, Optional, Tuple, TypeVar, Union

import attr
from attrs import define, field

from datumaro.components.cli_plugin import CliPlugin
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.errors import (
    AnnotationExportError,
    DatasetExportError,
    DatumaroError,
    ItemExportError,
)
from datumaro.components.media import Image, PointCloud
from datumaro.components.progress_reporting import NullProgressReporter, ProgressReporter
from datumaro.util.meta_file_util import save_meta_file
from datumaro.util.os_util import rmtree
from datumaro.util.scope import on_error_do, scoped

T = TypeVar("T")


class _ExportFail(DatumaroError):
    pass


class ExportErrorPolicy:
    def report_item_error(self, error: Exception, *, item_id: Tuple[str, str]) -> None:
        """
        Allows to report a problem with a dataset item.
        If this function returns, the converter must skip the item.
        """

        if not isinstance(error, _ExportFail):
            ie = ItemExportError(item_id)
            ie.__cause__ = error
            return self._handle_item_error(ie)
        else:
            raise error

    def report_annotation_error(self, error: Exception, *, item_id: Tuple[str, str]) -> None:
        """
        Allows to report a problem with a dataset item annotation.
        If this function returns, the converter must skip the annotation.
        """

        if not isinstance(error, _ExportFail):
            ie = AnnotationExportError(item_id)
            ie.__cause__ = error
            return self._handle_annotation_error(ie)
        else:
            raise error

    def _handle_item_error(self, error: ItemExportError) -> None:
        """This function must either call fail() or return."""
        self.fail(error)

    def _handle_annotation_error(self, error: AnnotationExportError) -> None:
        """This function must either call fail() or return."""
        self.fail(error)

    def fail(self, error: Exception) -> NoReturn:
        raise _ExportFail from error


class FailingExportErrorPolicy(ExportErrorPolicy):
    pass


@define(eq=False)
class ExportContext:
    progress_reporter: ProgressReporter = field(
        default=None, converter=attr.converters.default_if_none(factory=NullProgressReporter)
    )
    error_policy: ExportErrorPolicy = field(
        default=None, converter=attr.converters.default_if_none(factory=FailingExportErrorPolicy)
    )


class NullExportContext(ExportContext):
    pass


class Exporter(CliPlugin):
    DEFAULT_IMAGE_EXT = None

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)

        # Deprecated
        parser.add_argument(
            "--save-images",
            action="store_true",
            default=None,
            help="(Deprecated. Use --save-media instead) " "Save images (default: False)",
        )

        parser.add_argument(
            "--save-media",
            action="store_true",
            default=None,  # TODO: remove default once save-images is removed
            help="Save media (default: False)",
        )
        parser.add_argument(
            "--image-ext",
            default=None,
            help="Image extension (default: keep or use format default%s)"
            % (" " + cls.DEFAULT_IMAGE_EXT if cls.DEFAULT_IMAGE_EXT else ""),
        )
        parser.add_argument(
            "--save-dataset-meta",
            action="store_true",
            help="Save dataset meta file (default: %(default)s)",
        )

        return parser

    @classmethod
    def convert(cls, extractor, save_dir, **options):
        converter = cls(extractor, save_dir, **options)
        return converter.apply()

    @classmethod
    @scoped
    def patch(cls, dataset, patch, save_dir, **options):
        # This solution is not any better in performance than just
        # writing a dataset, but in case of patching (i.e. writing
        # to the previous location), it allows to avoid many problems
        # with removing and replacing existing files. Surely, this
        # approach also has problems with removal of the given directory.
        # Problems can occur if we can't remove the directory,
        # or want to reuse the given directory. It can happen if it
        # is mounted or (sym-)linked.
        # Probably, a better solution could be to wipe directory
        # contents and write new data there. Note that directly doing this
        # also doesn't work, because images may be needed for writing.

        if not osp.isdir(save_dir):
            return cls.convert(dataset, save_dir, **options)

        tmpdir = mkdtemp(dir=osp.dirname(save_dir), prefix=osp.basename(save_dir), suffix=".tmp")
        on_error_do(rmtree, tmpdir, ignore_errors=True)
        shutil.copymode(save_dir, tmpdir)

        retval = cls.convert(dataset, tmpdir, **options)

        rmtree(save_dir)
        os.replace(tmpdir, save_dir)

        return retval

    def apply(self):
        raise NotImplementedError("Should be implemented in a subclass")

    def __init__(
        self,
        extractor: IDataset,
        save_dir: str,
        *,
        save_images=None,  # Deprecated
        save_media: Optional[bool] = None,
        image_ext: Optional[str] = None,
        default_image_ext: Optional[str] = None,
        save_dataset_meta: bool = False,
        ctx: Optional[ExportContext] = None,
    ):
        default_image_ext = default_image_ext or self.DEFAULT_IMAGE_EXT
        assert default_image_ext
        self._default_image_ext = default_image_ext

        if save_images is not None and save_media is not None:
            raise DatasetExportError("Can't use both 'save-media' and " "'save-images'")

        if save_media is not None:
            self._save_media = save_media
        elif save_images is not None:
            self._save_media = save_images
            warnings.warn(
                "'save-images' is deprecated and will be "
                "removed in future. Use 'save-media' instead.",
                DeprecationWarning,
                stacklevel=2,
            )
        else:
            self._save_media = False

        self._image_ext = image_ext

        self._extractor = extractor
        self._save_dir = save_dir

        self._save_dataset_meta = save_dataset_meta

        # TODO: refactor this variable.
        # Can be used by a subclass to store the current patch info
        from datumaro.components.dataset import DatasetPatch

        if isinstance(extractor, DatasetPatch.DatasetPatchWrapper):
            self._patch = extractor.patch
        else:
            self._patch = None

        self._ctx: ExportContext = ctx or NullExportContext()

    def _find_image_ext(self, item: Union[DatasetItem, Image]):
        src_ext = None

        if isinstance(item, DatasetItem) and isinstance(item.media, Image):
            src_ext = item.media.ext
        elif isinstance(item, Image):
            src_ext = item.ext

        return self._image_ext or src_ext or self._default_image_ext

    def _make_item_filename(self, item, *, name=None, subdir=None):
        name = name or item.id
        subdir = subdir or ""
        return osp.join(subdir, name)

    def _make_image_filename(self, item, *, name=None, subdir=None):
        return self._make_item_filename(item, name=name, subdir=subdir) + self._find_image_ext(item)

    def _make_pcd_filename(self, item, *, name=None, subdir=None):
        return self._make_item_filename(item, name=name, subdir=subdir) + ".pcd"

    def _save_image(self, item, path=None, *, name=None, subdir=None, basedir=None):
        assert not (
            (subdir or name or basedir) and path
        ), "Can't use both subdir or name or basedir and path arguments"

        if not isinstance(item.media, Image) or not item.media.has_data:
            log.warning("Item '%s' has no image", item.id)
            return

        basedir = basedir or self._save_dir
        path = path or osp.join(basedir, self._make_image_filename(item, name=name, subdir=subdir))
        path = osp.abspath(path)

        item.media.save(path)

    def _save_point_cloud(self, item=None, path=None, *, name=None, subdir=None, basedir=None):
        assert not (
            (subdir or name or basedir) and path
        ), "Can't use both subdir or name or basedir and path arguments"

        if not item.media or not isinstance(item.media, PointCloud):
            log.warning("Item '%s' has no pcd", item.id)
            return

        basedir = basedir or self._save_dir
        path = path or osp.join(basedir, self._make_pcd_filename(item, name=name, subdir=subdir))
        path = osp.abspath(path)

        os.makedirs(osp.dirname(path), exist_ok=True)
        if item.media and osp.isfile(item.media.path):
            if item.media.path != path:
                shutil.copyfile(item.media.path, path)

    def _save_meta_file(self, path):
        save_meta_file(path, self._extractor.categories())



# Copyright (C) 2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

from __future__ import annotations

import math
from typing import Iterable, Optional, Tuple, TypeVar

T = TypeVar("T")


class ProgressReporter:
    """
    Only one set of methods must be called:
        - start - report_status - finish
        - iter
        - split

    This class is supposed to manage the state of children progress bars
    and release of their resources, if necessary.
    """

    @property
    def period(self) -> float:
        """
        Returns reporting period.

        For example, 0.1 would mean every 10%.
        """
        raise NotImplementedError

    def start(self, total: int, *, desc: Optional[str] = None):
        """Initializes the progress bar"""
        raise NotImplementedError

    def report_status(self, progress: int):
        """Updates the progress bar"""
        raise NotImplementedError

    def finish(self):
        """Finishes the progress bar"""
        pass  # pylint: disable=unnecessary-pass

    def iter(
        self, iterable: Iterable[T], *, total: Optional[int] = None, desc: Optional[str] = None
    ) -> Iterable[T]:
        """
        Traverses the iterable and reports progress simultaneously.

        Starts and finishes the progress bar automatically.

        Args:
            iterable: An iterable to be traversed
            total: The expected number of iterations. If not provided, will
              try to use iterable.__len__.
            desc: The status message

        Returns:
            An iterable over elements of the input sequence
        """

        if total is None and hasattr(iterable, "__len__"):
            total = len(iterable)

        self.start(total, desc=desc)

        if total:
            display_step = math.ceil(total * self.period)

        for i, elem in enumerate(iterable):
            if not total or i % display_step == 0:
                self.report_status(i)

            yield elem

        self.finish()

    def split(self, count: int) -> Tuple[ProgressReporter, ...]:
        """
        Splits the progress bar into few independent parts.
        In case of 0 must return an empty tuple.

        This class is supposed to manage the state of children progress bars
        and release of their resources, if necessary.
        """
        raise NotImplementedError


class NullProgressReporter(ProgressReporter):
    @property
    def period(self) -> float:
        return 0

    def start(self, total: int, *, desc: Optional[str] = None):
        pass

    def report_status(self, progress: int):
        pass

    def iter(
        self, iterable: Iterable[T], *, total: Optional[int] = None, desc: Optional[str] = None
    ) -> Iterable[T]:
        yield from iterable

    def split(self, count: int) -> Tuple[ProgressReporter]:
        return (self,) * count



# Copyright (C) 2020-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

import os
import os.path as osp
import re

from datumaro.components.dataset_base import DatasetBase, DatasetItem
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.importer import Importer
from datumaro.components.media import Image
from datumaro.util import str_to_bool
from datumaro.util.image import find_images


class Market1501Path:
    QUERY_DIR = "query"
    BBOX_DIR = "bounding_box_"
    IMAGE_EXT = ".jpg"
    PATTERN = re.compile(r"^(-?\d+)_c(\d+)s(\d+)_(\d+)_(\d+)(.*)")
    LIST_PREFIX = "images_"
    UNKNOWN_ID = -1
    ATTRIBUTES = ["person_id", "camera_id", "track_id", "frame_id", "bbox_id"]


class Market1501Base(DatasetBase):
    def __init__(self, path):
        if not osp.isdir(path):
            raise NotADirectoryError("Can't open folder with annotation files '%s'" % path)

        self._path = path
        super().__init__()

        subsets = {}
        for p in os.listdir(path):
            pf = osp.join(path, p)

            if p.startswith(Market1501Path.BBOX_DIR) and osp.isdir(pf):
                subset = p.replace(Market1501Path.BBOX_DIR, "")
                subsets[subset] = pf

            if p.startswith(Market1501Path.LIST_PREFIX) and osp.isfile(pf):
                subset = p.replace(Market1501Path.LIST_PREFIX, "")
                subset = osp.splitext(subset)[0]
                subsets[subset] = pf

            if p.startswith(Market1501Path.QUERY_DIR) and osp.isdir(pf):
                subset = Market1501Path.QUERY_DIR
                subsets[subset] = pf

        self._items = []
        for subset, subset_path in subsets.items():
            self._items.extend(list(self._load_items(subset, subset_path).values()))

    def __iter__(self):
        yield from self._items

    def _load_items(self, subset, subset_path):
        items = {}

        paths = []
        if osp.isfile(subset_path):
            with open(subset_path, encoding="utf-8") as f:
                for line in f:
                    paths.append(osp.join(self._path, line.strip()))
        else:
            paths = list(find_images(subset_path, recursive=True))

        for image_path in sorted(paths):
            item_id = osp.splitext(osp.normpath(image_path))[0]
            if osp.isabs(image_path):
                item_id = osp.relpath(item_id, self._path)
            item_id = item_id.split(osp.sep, maxsplit=1)[1]

            attributes = {}
            search = Market1501Path.PATTERN.search(osp.basename(item_id))
            if search:
                attribute_values = search.groups()[0:5]
                attributes = {
                    "person_id": attribute_values[0],
                    "camera_id": int(attribute_values[1]) - 1,
                    "track_id": int(attribute_values[2]),
                    "frame_id": int(attribute_values[3]),
                    "bbox_id": int(attribute_values[4]),
                    "query": subset == Market1501Path.QUERY_DIR,
                }

                custom_name = search.groups()[5]
                if custom_name:
                    item_id = osp.join(osp.dirname(item_id), custom_name)

            item = items.get(item_id)
            if item is None:
                item = DatasetItem(
                    id=item_id, subset=subset, media=Image(path=image_path), attributes=attributes
                )
                items[item_id] = item

        return items


class Market1501Importer(Importer):
    @classmethod
    def find_sources(cls, path):
        for dirname in os.listdir(path):
            if dirname.startswith(
                (Market1501Path.BBOX_DIR, Market1501Path.QUERY_DIR, Market1501Path.LIST_PREFIX)
            ):
                return [{"url": path, "format": Market1501Base.NAME}]


class Market1501Exporter(Exporter):
    DEFAULT_IMAGE_EXT = Market1501Path.IMAGE_EXT

    def _make_dir_name(self, item):
        dirname = Market1501Path.BBOX_DIR + item.subset
        query = item.attributes.get("query")
        if query is not None and isinstance(query, str):
            query = str_to_bool(query)
        if query:
            dirname = Market1501Path.QUERY_DIR
        return dirname

    def apply(self):
        if self._extractor.media_type() and not issubclass(self._extractor.media_type(), Image):
            raise MediaTypeError("Media type is not an image")

        for subset_name, subset in self._extractor.subsets().items():
            annotation = ""
            used_frames = {}

            for item in subset:
                dirname = self._make_dir_name(item)

                image_name = item.id
                pid = item.attributes.get("person_id")
                match = Market1501Path.PATTERN.fullmatch(item.id)
                if not match and pid:
                    cid = int(item.attributes.get("camera_id", 0)) + 1
                    tid = int(item.attributes.get("track_id", 1))
                    bbid = int(item.attributes.get("bbox_id", 0))
                    fid = int(
                        item.attributes.get(
                            "frame_id", max(used_frames.get((pid, cid, tid), [-1])) + 1
                        )
                    )
                    image_name = osp.join(
                        osp.dirname(image_name), f"{pid}_c{cid}s{tid}_{fid:06d}_{bbid:02d}"
                    )

                image_path = self._make_image_filename(item, name=image_name, subdir=dirname)
                if self._save_media and item.media:
                    self._save_image(item, osp.join(self._save_dir, image_path))

                attrs = Market1501Path.PATTERN.search(image_name)
                if attrs:
                    attrs = attrs.groups()
                    used_frames.setdefault(attrs[0:2], []).append(int(attrs[3]))
                annotation += "%s\n" % image_path

            annotation_file = osp.join(
                self._save_dir, Market1501Path.LIST_PREFIX + subset_name + ".txt"
            )
            with open(annotation_file, "w", encoding="utf-8") as f:
                f.write(annotation)



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import logging as log
import os
import os.path as osp
import warnings
from collections import Counter
from enum import Enum, auto
from itertools import zip_longest
from typing import Union

import cv2
import numpy as np

from datumaro.components.media import Image

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    import tensorboardX as tb

from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.dataset import IDataset
from datumaro.util import parse_str_enum_value
from datumaro.util.image import save_image


class DiffVisualizer:
    class OutputFormat(Enum):
        simple = auto()
        tensorboard = auto()

    DEFAULT_FORMAT = OutputFormat.simple
    _UNMATCHED_LABEL = -1

    def __init__(
        self, comparator, save_dir: str, output_format: Union[None, str, OutputFormat] = None
    ):
        self._cmp = comparator

        self._output_format = parse_str_enum_value(
            output_format, self.OutputFormat, default=self.DEFAULT_FORMAT
        )

        self._save_dir = save_dir

    def __enter__(self):
        os.makedirs(self._save_dir, exist_ok=True)

        if self._output_format is self.OutputFormat.tensorboard:
            logdir = osp.join(self._save_dir, "logs", "diff")
            self._file_writer = tb.SummaryWriter(logdir)
        elif self._output_format is self.OutputFormat.simple:
            self._label_diff_writer = None

        self._a_classes = {}
        self._b_classes = {}

        self.label_confusion_matrix = Counter()
        self.bbox_confusion_matrix = Counter()
        self.polygon_confusion_matrix = Counter()
        self.mask_confusion_matrix = Counter()

        return self

    def __exit__(self, *args, **kwargs):
        if self._output_format is self.OutputFormat.tensorboard:
            self._file_writer.flush()
            self._file_writer.close()
        elif self._output_format is self.OutputFormat.simple:
            if self._label_diff_writer:
                self._label_diff_writer.flush()
                self._label_diff_writer.close()

    def save(self, a: IDataset, b: IDataset):
        if len(a) != len(b):
            print("Datasets have different lengths: %s vs %s" % (len(a), len(b)))

        a_classes = a.categories().get(AnnotationType.label, LabelCategories())
        b_classes = b.categories().get(AnnotationType.label, LabelCategories())
        class_mismatch = [
            (idx, a_cls, b_cls)
            for idx, (a_cls, b_cls) in enumerate(zip_longest(a_classes, b_classes))
            if getattr(a_cls, "name", None) != getattr(b_cls, "name", None)
        ]
        if class_mismatch:
            print("Datasets have mismatching labels:")
            for idx, a_class, b_class in class_mismatch:
                if a_class and b_class:
                    print("  #%s: %s != %s" % (idx, a_class.name, b_class.name))
                elif a_class:
                    print("  #%s:  > %s" % (idx, a_class.name))
                else:
                    print("  #%s:  < %s" % (idx, b_class.name))
        self._a_classes = a.categories().get(AnnotationType.label)
        self._b_classes = b.categories().get(AnnotationType.label)

        ids_a = set((item.id, item.subset) for item in a)
        ids_b = set((item.id, item.subset) for item in b)
        ids = ids_a & ids_b

        if len(ids) != len(ids_a):
            print("Unmatched items in the first dataset: ")
            print(ids_a - ids)
        if len(ids) != len(ids_b):
            print("Unmatched items in the second dataset: ")
            print(ids_b - ids)

        for item_id, item_subset in ids:
            item_a = a.get(item_id, item_subset)
            item_b = b.get(item_id, item_subset)

            label_diff = self._cmp.match_labels(item_a, item_b)
            self.update_label_confusion(label_diff)

            bbox_diff = self._cmp.match_boxes(item_a, item_b)
            self.update_bbox_confusion(bbox_diff)

            polygon_diff = self._cmp.match_polygons(item_a, item_b)
            self.update_polygon_confusion(polygon_diff)

            mask_diff = self._cmp.match_masks(item_a, item_b)
            self.update_mask_confusion(mask_diff)

            self.save_item_label_diff(item_a, item_b, label_diff)

            if (
                a.media_type()
                and issubclass(a.media_type(), Image)
                and b.media_type()
                and issubclass(b.media_type(), Image)
            ):
                self.save_item_bbox_diff(item_a, item_b, bbox_diff)

        if len(self.label_confusion_matrix) != 0:
            self.save_conf_matrix(self.label_confusion_matrix, "label_confusion.png")
        if len(self.bbox_confusion_matrix) != 0:
            self.save_conf_matrix(self.bbox_confusion_matrix, "bbox_confusion.png")
        if len(self.polygon_confusion_matrix) != 0:
            self.save_conf_matrix(self.polygon_confusion_matrix, "polygon_confusion.png")
        if len(self.mask_confusion_matrix) != 0:
            self.save_conf_matrix(self.mask_confusion_matrix, "mask_confusion.png")

    def update_label_confusion(self, label_diff):
        matches, a_unmatched, b_unmatched = label_diff
        for label in matches:
            self.label_confusion_matrix[(label, label)] += 1
        for a_label in a_unmatched:
            self.label_confusion_matrix[(a_label, self._UNMATCHED_LABEL)] += 1
        for b_label in b_unmatched:
            self.label_confusion_matrix[(self._UNMATCHED_LABEL, b_label)] += 1

    @classmethod
    def _update_segment_confusion(cls, matrix, diff):
        matches, mispred, a_unmatched, b_unmatched = diff
        for a_segm, b_segm in matches:
            matrix[(a_segm.label, b_segm.label)] += 1
        for a_segm, b_segm in mispred:
            matrix[(a_segm.label, b_segm.label)] += 1
        for a_segm in a_unmatched:
            matrix[(a_segm.label, cls._UNMATCHED_LABEL)] += 1
        for b_segm in b_unmatched:
            matrix[(cls._UNMATCHED_LABEL, b_segm.label)] += 1

    def update_bbox_confusion(self, diff):
        self._update_segment_confusion(self.bbox_confusion_matrix, diff)

    def update_polygon_confusion(self, diff):
        self._update_segment_confusion(self.polygon_confusion_matrix, diff)

    def update_mask_confusion(self, diff):
        self._update_segment_confusion(self.mask_confusion_matrix, diff)

    @classmethod
    def draw_text_with_background(
        cls,
        frame,
        text,
        origin,
        font=None,
        scale=1.0,
        color=(0, 0, 0),
        thickness=1,
        bgcolor=(1, 1, 1),
    ):
        if not font:
            font = cv2.FONT_HERSHEY_SIMPLEX

        text_size, baseline = cv2.getTextSize(text, font, scale, thickness)
        cv2.rectangle(
            frame,
            tuple((origin + (0, baseline)).astype(int)),
            tuple((origin + (text_size[0], -text_size[1])).astype(int)),
            bgcolor,
            cv2.FILLED,
        )
        cv2.putText(frame, text, tuple(origin.astype(int)), font, scale, color, thickness)
        return text_size, baseline

    def draw_detection_roi(self, frame, x, y, w, h, label, conf, color):
        cv2.rectangle(frame, (x, y), (x + w, y + h), color, 2)

        text = "%s %.2f%%" % (label, 100.0 * conf)
        text_scale = 0.5
        font = cv2.FONT_HERSHEY_SIMPLEX
        text_size = cv2.getTextSize(text, font, text_scale, 1)
        line_height = np.array([0, text_size[0][1]])
        self.draw_text_with_background(
            frame,
            text,
            np.array([x, y]) - line_height * 0.5,
            font,
            scale=text_scale,
            color=[255 - c for c in color],
        )

    def get_a_label(self, label_id):
        return self._get_label(self._a_classes, label_id)

    def get_b_label(self, label_id):
        return self._get_label(self._b_classes, label_id)

    @staticmethod
    def _get_label(cat: LabelCategories, label_id):
        if cat is None:
            return str(label_id)
        return cat[label_id].name

    def draw_bbox(self, img, shape, label, color):
        x, y, w, h = shape.get_bbox()
        self.draw_detection_roi(
            img, int(x), int(y), int(w), int(h), label, shape.attributes.get("score", 1), color
        )

    def get_label_diff_file(self):
        if self._label_diff_writer is None:
            self._label_diff_writer = open(
                osp.join(self._save_dir, "label_diff.txt"), "w", encoding="utf-8"
            )
        return self._label_diff_writer

    def save_item_label_diff(self, item_a, item_b, diff):
        _, a_unmatched, b_unmatched = diff

        if 0 < len(a_unmatched) + len(b_unmatched):
            if self._output_format is self.OutputFormat.simple:
                f = self.get_label_diff_file()
                f.write(item_a.id + "\n")
                for a_label in a_unmatched:
                    f.write("  >%s\n" % self.get_a_label(a_label))
                for b_label in b_unmatched:
                    f.write("  <%s\n" % self.get_b_label(b_label))
            elif self._output_format is self.OutputFormat.tensorboard:
                tag = item_a.id
                for a_label in a_unmatched:
                    self._file_writer.add_text(tag, ">%s\n" % self.get_a_label(a_label))
                for b_label in b_unmatched:
                    self._file_writer.add_text(tag, "<%s\n" % self.get_b_label(b_label))

    def save_item_bbox_diff(self, item_a, item_b, diff):
        _, mispred, a_unmatched, b_unmatched = diff

        if 0 < len(a_unmatched) + len(b_unmatched) + len(mispred):
            if not isinstance(item_a.media, Image) or not item_a.media.has_data:
                log.warning("Item %s: item has no image data, " "it will be skipped" % (item_a.id))
                return
            img_a = item_a.media.data.copy()
            img_b = img_a.copy()
            for a_bbox, b_bbox in mispred:
                self.draw_bbox(img_a, a_bbox, self.get_a_label(a_bbox.label), (0, 255, 0))
                self.draw_bbox(img_b, b_bbox, self.get_b_label(b_bbox.label), (0, 0, 255))
            for a_bbox in a_unmatched:
                self.draw_bbox(img_a, a_bbox, self.get_a_label(a_bbox.label), (255, 255, 0))
            for b_bbox in b_unmatched:
                self.draw_bbox(img_b, b_bbox, self.get_b_label(b_bbox.label), (255, 255, 0))

            img = np.hstack([img_a, img_b])

            path = osp.join(self._save_dir, item_a.subset, item_a.id)

            if self._output_format is self.OutputFormat.simple:
                save_image(path + ".png", img, create_dir=True)
            elif self._output_format is self.OutputFormat.tensorboard:
                self.save_as_tensorboard(img, path)

    def save_as_tensorboard(self, img, name):
        img = img[:, :, ::-1]  # to RGB
        img = np.transpose(img, (2, 0, 1))  # to (C, H, W)
        img = img.astype(dtype=np.uint8)
        self._file_writer.add_image(name, img)

    def save_conf_matrix(self, conf_matrix, filename):
        import matplotlib.pyplot as plt

        def _get_class_map(label_categories):
            classes = None
            if label_categories is not None:
                classes = {id: c.name for id, c in enumerate(label_categories.items)}
            if classes is None:
                classes = {c: "label_%s" % c for c, _ in conf_matrix}
            classes[self._UNMATCHED_LABEL] = "unmatched"
            classes[None] = "no_class"
            return classes

        a_classes = _get_class_map(self._a_classes)
        b_classes = _get_class_map(self._b_classes)

        a_class_idx = {id: i for i, id in enumerate(a_classes)}
        b_class_idx = {id: i for i, id in enumerate(b_classes)}
        matrix = np.zeros((len(a_classes), len(b_classes)), dtype=int)
        for idx_pair in conf_matrix:
            index = (a_class_idx[idx_pair[0]], b_class_idx[idx_pair[1]])
            matrix[index] = conf_matrix[idx_pair]

        a_labels = [label for id, label in a_classes.items()]
        b_labels = [label for id, label in b_classes.items()]

        fig = plt.figure()
        fig.add_subplot(111)
        table = plt.table(cellText=matrix, rowLabels=a_labels, colLabels=b_labels, loc="center")
        table.auto_set_font_size(False)
        table.set_fontsize(8)
        table.scale(3, 3)
        # Removing ticks and spines enables you to get the figure only with table
        plt.tick_params(axis="x", which="both", bottom=False, top=False, labelbottom=False)
        plt.tick_params(axis="y", which="both", right=False, left=False, labelleft=False)
        for pos in ["right", "top", "bottom", "left"]:
            plt.gca().spines[pos].set_visible(False)

        for idx_pair in conf_matrix:
            i = a_class_idx[idx_pair[0]]
            j = b_class_idx[idx_pair[1]]
            if conf_matrix[idx_pair] != 0:
                if a_classes[idx_pair[0]] == b_classes[idx_pair[1]]:
                    table._cells[(i + 1, j)].set_facecolor("#00FF00")
                else:
                    table._cells[(i + 1, j)].set_facecolor("#FF0000")

        plt.savefig(osp.join(self._save_dir, filename), bbox_inches="tight", pad_inches=0.05)



# Copyright (C) 2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse

from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.project import load_project


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Navigate to a revision",
        description="""
        Command forms:|n
        1) %(prog)s <revision>|n
        2) %(prog)s [--] <source1> ...|n
        3) %(prog)s <revision> [--] <source1> <source2> ...|n
        |n
        1 - Restores a revision and all the sources in the working directory.|n
        2, 3 - Restores only specified sources from the specified revision.|n
        |n
        The current revision is used, when not set. When no revision
        and no sources is specified, restores the
        current revision (like the 1-st form).
        "--" is optionally used to separate source names and revisions.|n
        |n
        Examples:|n
        - Restore the previous revision:|n
        |s|s%(prog)s HEAD~1 |n |n
        - Restore the saved version of a source in the working tree|n
        |s|s%(prog)s -- source-1 |n |n
        - Restore a previous version of a source|n
        |s|s%(prog)s 33fbfbe my-source
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "_positionals", nargs=argparse.REMAINDER, help=argparse.SUPPRESS
    )  # workaround for -- eaten by positionals
    parser.add_argument("rev", nargs="?", help="Commit or tag (default: current)")
    parser.add_argument("sources", nargs="*", help="Sources to restore (default: all)")
    parser.add_argument(
        "-f",
        "--force",
        action="store_true",
        help="Allows to overwrite unsaved changes in case of conflicts " "(default: %(default)s)",
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=checkout_command)

    return parser


def get_sensitive_args():
    return {
        checkout_command: [
            "rev",
            "sources",
            "project_dir",
        ],
    }


@scoped
def checkout_command(args):
    has_sep = "--" in args._positionals
    if has_sep:
        pos = args._positionals.index("--")
        if 1 < pos:
            raise argparse.ArgumentError(None, message="Expected no more than 1 revision argument")
    else:
        pos = 1
    args.rev = (args._positionals[:pos] or [None])[0]
    args.sources = args._positionals[pos + has_sep :]
    if has_sep and not args.sources:
        raise argparse.ArgumentError(
            "sources", message="When '--' is used, " "at least 1 source name must be specified"
        )

    project = scope_add(load_project(args.project_dir))

    project.checkout(rev=args.rev, sources=args.sources, force=args.force)

    return 0



# Copyright (C) 2022 Intel Corporation
#
# SPDX-License-Identifier: MIT
import math
import warnings
from collections import defaultdict
from typing import Iterable, List, Optional, Tuple, Union

import cv2
import matplotlib.patches as patches
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.text import Text
from mpl_toolkits.axes_grid1 import make_axes_locatable
from PIL import ImageColor

from datumaro.components.annotation import (
    Annotation,
    AnnotationType,
    Bbox,
    Caption,
    Cuboid3d,
    DepthAnnotation,
    Label,
    LabelCategories,
    Mask,
    Points,
    Polygon,
    PolyLine,
    SuperResolutionAnnotation,
)
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.media import Image

CAPTION_BBOX_PAD = 0.2
DEFAULT_COLOR_CYCLES: List[str] = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]


def _infer_grid_size(length: int, grid_size: Tuple[Optional[int], Optional[int]]):
    nrows, ncols = grid_size

    if nrows is None and ncols is None:
        nrows = ncols = int(math.sqrt(length))

        while nrows * ncols < length:
            nrows += 1
    elif nrows is None and ncols > 0:
        nrows = int(length / ncols)

        while nrows * ncols < length:
            nrows += 1
    elif nrows > 0 and ncols is None:
        ncols = int(length / nrows)

        while nrows * ncols < length:
            ncols += 1

    assert nrows > 0, "nrows should be a positive integer."
    assert ncols > 0, "ncols should be a positive integer."
    assert length <= nrows * ncols, "The number of ids should less then or equal to nrows * ncols."

    return nrows, ncols


class Visualizer:
    def __init__(
        self,
        dataset: IDataset,
        ignored_types: Optional[Iterable[AnnotationType]] = None,
        figsize: Tuple[float, float] = (8, 6),
        color_cycles: Optional[List[str]] = None,
        bbox_linewidth: float = 1.0,
        text_y_offset: float = 1.5,
        alpha: float = 1.0,
    ) -> None:
        """
        Visualizer for Datumaro annotations

        Parameters
        ----------
        dataset:
            Datumaro dataset to visualize its items.
        ignored_types:
            Categories of labels. It is used to extract label name by label id.
        figsize:
            Pyplot Figure instance used to draw annotation.
        color_cycles:
            Color cycle corresponding to each label ID.
            If the length of the color cycle is less than the label ID,
            then the label ID exceeding the color cycle length is assgined by the following rule.
            color = color_cycles[label_id % len(color_cycles)]
        bbox_linewidth:
            Line width for Bbox, Polygon and PolyLine annotation
        text_y_offset:
            Offset of y axis for texts.
            The higher value puts the text in the upper place of the annotation.
        alpha:
            Transparency value when drawing annotations. It should be in [0, 1].
            If alpha=0, we do not draw any annotations.
        """
        self.dataset = dataset
        self.figsize = figsize
        self.ignored_types = set(ignored_types) if ignored_types is not None else set()
        self.color_cycles = color_cycles if color_cycles is not None else DEFAULT_COLOR_CYCLES
        self.bbox_linewidth = bbox_linewidth
        self.text_y_offset = text_y_offset

        assert 0.0 <= alpha <= 1.0, "alpha should be in [0, 1]."
        self.alpha = alpha

    @property
    def draw_only_image(self):
        """
        If self.alpha = 0, we do not overdraw any annotation over the image.
        """
        return self.alpha == 0.0

    def _draw(
        self,
        ann: Annotation,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        """
        Draw annotation according to it's annotation type.

        Parameters
        ----------
        ann:
            Annotation entity to draw.
        label_categories:
            Categories of labels. It is used to extract label name by label id.
        fig:
            Pyplot Figure instance used to draw annotation.
        ax:
            Pyplot Axes instance used to draw annotation.
        context:
            It includes previously drawing history for each annotation type.
            Currently, it is necessary to avoid drawing again over an already drawn place.
            For example, multi label dataset has multiple Labels in a DatasetItem.
            If we don't keep context, it has to draw Label at the same place again and again.
        """
        if ann.type == AnnotationType.label:
            return self._draw_label(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.mask:
            return self._draw_mask(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.points:
            return self._draw_points(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.polygon:
            return self._draw_polygon(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.polyline:
            return self._draw_polygon(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.bbox:
            return self._draw_bbox(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.caption:
            return self._draw_caption(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.cuboid_3d:
            return self._draw_cuboid_3d(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.super_resolution_annotation:
            return self._draw_super_resolution_annotation(ann, label_categories, fig, ax, context)
        if ann.type == AnnotationType.depth_annotation:
            return self._draw_depth_annotation(ann, label_categories, fig, ax, context)

        raise ValueError(f"Unknown ann.type={ann.type}")

    def _get_color(self, ann: Annotation) -> str:
        color = self.color_cycles[ann.label % len(self.color_cycles)]
        return color

    def _sort_by_z_order(self, annotations: List[Annotation]) -> List[Annotation]:
        def _sort_key(ann: Annotation):
            z_order = getattr(ann, "z_order", -1)
            return z_order

        return sorted(annotations, key=_sort_key)

    def vis_gallery(
        self,
        ids: List[Union[str, DatasetItem]],
        subset: Optional[Union[str, List[str]]] = None,
        grid_size: Tuple[Optional[int], Optional[int]] = (None, None),
    ) -> Figure:
        nrows, ncols = _infer_grid_size(len(ids), grid_size)
        fig, axs = plt.subplots(nrows, ncols, figsize=self.figsize)

        if isinstance(subset, list):
            assert len(ids) == len(
                subset
            ), "If subset is a list, it should have the same length as ids."

        for i, (dataset_id, ax) in enumerate(zip(ids, axs.flatten())):
            if isinstance(subset, List):
                self.vis_one_sample(dataset_id, subset[i], ax)
            else:
                self.vis_one_sample(dataset_id, subset, ax)

        return fig

    def vis_one_sample(
        self, id: Union[str, DatasetItem], subset: Optional[str] = None, ax: Optional[Axes] = None
    ) -> Figure:
        if ax is None:
            fig = plt.figure(figsize=self.figsize)
            ax = plt.gca()
        else:
            fig = ax.get_figure()
            plt.sca(ax)

        item: DatasetItem = self.dataset.get(id, subset)
        assert item is not None, f"Cannot find id={id}, subset={subset}"
        assert (
            item is not Image
        ), f"Media type should be Image, Current media type={type(item.media)}"

        img = item.media.data.astype(np.uint8)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        ax.imshow(img)

        width = ax.transAxes.transform_point((1, 0))[0] - ax.transAxes.transform_point((0, 0))[0]
        text = ax.set_title(f"ID: {id}, Subset: {subset}", loc="center", wrap=True)
        text.__get_wrapped_text = text._get_wrapped_text

        def _get_wrapped_text():
            wrapped_text = text.__get_wrapped_text()
            text._text = wrapped_text
            return wrapped_text

        text._get_wrapped_text = _get_wrapped_text
        text._get_wrap_line_width = lambda: width

        ax.set_axis_off()

        if self.draw_only_image:
            return fig

        annotations = self._sort_by_z_order(item.annotations)
        categories = self.dataset.categories()
        label_categories = (
            self.dataset.categories()[AnnotationType.label]
            if AnnotationType.label in categories
            else None
        )

        context = defaultdict(list)
        for ann in annotations:
            if ann.type in self.ignored_types:
                warnings.warn(f"{ann.type} in self.ignored_types. Skip it.")
                continue
            self._draw(ann, label_categories, fig, ax, context[ann.type])

        return fig

    def _draw_label(
        self,
        ann: Label,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        label_text = label_categories[ann.label].name if label_categories is not None else ann.label
        color = self._get_color(ann)

        if len(context) == 0:
            x, y = 0.01, 0.99
        else:
            # Draw below the previously drawn label.
            text: Text = context[-1]
            # We can know the position of text bbox only if drawing it actually.
            # https://stackoverflow.com/a/41271773/16880031
            fig.canvas.draw()
            bbox = text.get_window_extent()
            bbox = ax.transAxes.inverted().transform_bbox(bbox)
            x, y = 0.01, bbox.y0

        text = ax.text(x, y, label_text, ha="left", va="top", color=color, transform=ax.transAxes)
        context.append(text)

    def _draw_mask(
        self,
        ann: Mask,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        h, w = ann.image.shape
        mask_map = np.zeros((h, w, 4), dtype=np.uint8)
        color = self._get_color(ann)
        rgba_color = (*ImageColor.getcolor(color, "RGB"), 0.0)
        mask_map[ann.image] = rgba_color
        mask_map[ann.image, 3] = int(255 * self.alpha)

        ax.imshow(mask_map)

    def _draw_points(
        self,
        ann: Points,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        label_text = label_categories[ann.label].name if label_categories is not None else ann.label
        color = self._get_color(ann)
        points = np.array(ann.points)
        n_points = len(points) // 2
        points = points.reshape(n_points, 2)
        visible = [viz == Points.Visibility.visible for viz in ann.visibility]
        points = points[visible]

        ax.scatter(points[:, 0], points[:, 1], color=color)

        if len(points) > 0:
            x, y, _, _ = ann.get_bbox()
            ax.text(x, y - self.text_y_offset, label_text, color=color)

    def _draw_polygon(
        self,
        ann: Union[Polygon, PolyLine],
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        label_text = label_categories[ann.label].name if label_categories is not None else ann.label
        color = self._get_color(ann)
        points = np.array(ann.points)
        n_points = len(points) // 2
        points = points.reshape(n_points, 2)

        polyline = patches.Polygon(
            points,
            fill=False,
            linewidth=self.bbox_linewidth,
            edgecolor=color,
        )
        ax.add_patch(polyline)

        if isinstance(ann, Polygon):
            polygon = patches.Polygon(
                points,
                fill=True,
                facecolor=color if isinstance(ann, Polygon) else "none",
                alpha=self.alpha,
            )
            ax.add_patch(polygon)

        x, y, _, _ = ann.get_bbox()
        ax.text(x, y - self.text_y_offset, label_text, color=color)

    def _draw_bbox(
        self,
        ann: Bbox,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        label_text = label_categories[ann.label].name if label_categories is not None else ann.label
        color = self._get_color(ann)
        rect = patches.Rectangle(
            (ann.x, ann.y),
            ann.w,
            ann.h,
            linewidth=self.bbox_linewidth,
            edgecolor=color,
            facecolor="none",
        )
        ax.text(ann.x, ann.y - self.text_y_offset, label_text, color=color)
        ax.add_patch(rect)

    def _draw_caption(
        self,
        ann: Caption,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        if len(context) == 0:
            x, y = 0.5, 0.01
        else:
            # Draw on the top of the previously drawn caption.
            text: Text = context[-1]
            bbox = text.get_bbox_patch()
            # We can know the position of text bbox only if drawing it actually.
            # https://stackoverflow.com/a/41271773/16880031
            fig.canvas.draw()
            drawed_bbox = bbox.get_bbox()
            x, y = bbox.get_transform().transform(
                [0, (1.0 + 2 * CAPTION_BBOX_PAD) * drawed_bbox.height]
            )
            x, y = ax.transAxes.inverted().transform([x, y])
            x, y = 0.5, y

        width = ax.transAxes.transform_point((1, 0))[0] - ax.transAxes.transform_point((0, 0))[0]

        text = ax.text(
            x,
            y,
            ann.caption,
            ha="center",
            va="bottom",
            wrap=True,
            transform=ax.transAxes,
            bbox={"facecolor": "white", "alpha": self.alpha},
        )
        text.get_bbox_patch().set_boxstyle(f"Round,pad={CAPTION_BBOX_PAD}")
        text._get_wrap_line_width = lambda: width
        context.append(text)

    def _draw_cuboid_3d(
        self,
        ann: Cuboid3d,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        raise NotImplementedError(f"{ann.type} is not implemented yet.")

    def _draw_super_resolution_annotation(
        self,
        ann: SuperResolutionAnnotation,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        assert (
            len(context) == 0
        ), "It cannot visualize more than one SuperResolutionAnnotation per item."

        warnings.warn(
            "SuperResolutionAnnotation overdraws the high-resolution image over the original image. "
            "If you want to see the original image, set alpha=0."
        )

        hi_res_img = ann.image.data
        im = ax.imshow(hi_res_img, alpha=self.alpha)
        context.append(im)

    def _draw_depth_annotation(
        self,
        ann: DepthAnnotation,
        label_categories: Optional[LabelCategories],
        fig: Figure,
        ax: Axes,
        context: List,
    ) -> None:
        assert len(context) == 0, "It cannot visualize more than one DepthAnnotation per item."

        depth = ann.image.data

        im = ax.imshow(depth, alpha=self.alpha)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)

        fig.colorbar(im, cax)
        context.append(im)



# Copyright (C) 2019-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log
import os
import os.path as osp

from datumaro.util.image import is_image, load_image, save_image
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.errors import CliException
from ..util.project import load_project, parse_full_revpath


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Run Explainable AI algorithm",
        description="""
        Runs an explainable AI algorithm for a model.|n
        |n
        This tool is supposed to help an AI developer to debug
        a model and a dataset. Basically, it executes inference and
        tries to find problems in the trained model - determine decision
        boundaries and belief intervals for the classifier.|n
        |n
        Currently, the only available algorithm is
        RISE (https://arxiv.org/pdf/1806.07421.pdf), which runs
        inference and then re-runs a model multiple times
        on each image to produce a heatmap of activations for
        each output of the first inference. As a result, we obtain
        few heatmaps, which shows, how image pixels affected
        the inference result. This algorithm doesn't require any special
        information about the model, but it requires the model to
        return all the outputs and confidences. Check the User Manual
        for usage examples.|n
        Supported scenarios:|n
        - RISE for classification|n
        - RISE for Object Detection|n
        |n
        This command has the following syntax:|n
        |s|s%(prog)s <image path or revpath>|n
        |n
        <image path> - a path to the file.|n
        <revpath> - either a dataset path or a revision path. The full
        syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        Parts can be enclosed in quotes.|n
        |n
        The current project (-p/--project) is used as a context for plugins
        and models. It is used when there is a dataset path in target.
        When not specified, the current project's working tree is used.|n
        |n
        Examples:|n
        - Run RISE on an image, display results:|n
        |s|s%(prog)s path/to/image.jpg -m mymodel rise --max-samples 50|n
        |n
        - Run RISE on a source revision:|n
        |s|s%(prog)s HEAD~1:source-1 -m model rise
        """,
        formatter_class=MultilineFormatter,
    )

    parser.add_argument(
        "target",
        nargs="?",
        default=None,
        help="Inference target - image, revpath (default: project)",
    )
    parser.add_argument("-m", "--model", required=True, help="Model to use for inference")
    parser.add_argument(
        "-o",
        "--output-dir",
        dest="save_dir",
        default=None,
        help="Directory to save output (default: display only)",
    )

    method_sp = parser.add_subparsers(dest="algorithm")

    rise_parser = method_sp.add_parser(
        "rise",
        description="""
        RISE: Randomized Input Sampling for
        Explanation of Black-box Models algorithm|n
        |n
        See explanations at: https://arxiv.org/pdf/1806.07421.pdf
        """,
        formatter_class=MultilineFormatter,
    )
    rise_parser.add_argument(
        "-s",
        "--max-samples",
        default=None,
        type=int,
        help="Number of algorithm iterations (default: mask size ^ 2)",
    )
    rise_parser.add_argument(
        "--mw",
        "--mask-width",
        dest="mask_width",
        default=7,
        type=int,
        help="Mask width (default: %(default)s)",
    )
    rise_parser.add_argument(
        "--mh",
        "--mask-height",
        dest="mask_height",
        default=7,
        type=int,
        help="Mask height (default: %(default)s)",
    )
    rise_parser.add_argument(
        "--prob",
        default=0.5,
        type=float,
        help="Mask pixel inclusion probability (default: %(default)s)",
    )
    rise_parser.add_argument(
        "--iou",
        "--iou-thresh",
        dest="iou_thresh",
        default=0.9,
        type=float,
        help="IoU match threshold for detections (default: %(default)s)",
    )
    rise_parser.add_argument(
        "--nms",
        "--nms-iou-thresh",
        dest="nms_iou_thresh",
        default=0.0,
        type=float,
        help="IoU match threshold in Non-maxima suppression (default: no NMS)",
    )
    rise_parser.add_argument(
        "--conf",
        "--det-conf-thresh",
        dest="det_conf_thresh",
        default=0.0,
        type=float,
        help="Confidence threshold for detections (default: include all)",
    )
    rise_parser.add_argument(
        "-b",
        "--batch-size",
        default=1,
        type=int,
        help="Inference batch size (default: %(default)s)",
    )
    rise_parser.add_argument(
        "--display", action="store_true", help="Visualize results during computations"
    )

    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to operate on (default: current dir)",
    )
    parser.set_defaults(command=explain_command)

    return parser


def get_sensitive_args():
    return {
        explain_command: [
            "target",
            "model",
            "save_dir",
            "project_dir",
        ],
    }


@scoped
def explain_command(args):
    import cv2
    from matplotlib import cm

    project = scope_add(load_project(args.project_dir))

    model = project.working_tree.models.make_executable_model(args.model)

    if str(args.algorithm).lower() != "rise":
        raise CliException("Explanation algorithm except for rise is not yet implemented.")

    from datumaro.components.algorithms.rise import RISE

    rise = RISE(
        model,
        max_samples=args.max_samples,
        mask_width=args.mask_width,
        mask_height=args.mask_height,
        prob=args.prob,
        iou_thresh=args.iou_thresh,
        nms_thresh=args.nms_iou_thresh,
        det_conf_thresh=args.det_conf_thresh,
        batch_size=args.batch_size,
    )

    if args.target and is_image(args.target):
        image_path = args.target
        image = load_image(image_path)

        log.info("Running inference explanation for '%s'" % image_path)
        heatmap_iter = rise.apply(image, progressive=args.display)

        image = image / 255.0
        file_name = osp.splitext(osp.basename(image_path))[0]
        if args.display:
            for i, heatmaps in enumerate(heatmap_iter):
                for j, heatmap in enumerate(heatmaps):
                    hm_painted = cm.jet(heatmap)[:, :, 2::-1]
                    disp = (image + hm_painted) / 2
                    cv2.imshow("heatmap-%s" % j, hm_painted)
                    cv2.imshow(file_name + "-heatmap-%s" % j, disp)
                cv2.waitKey(10)
                print("Iter", i, "of", args.max_samples, end="\r")
        else:
            heatmaps = next(heatmap_iter)

        if args.save_dir is not None:
            log.info("Saving inference heatmaps at '%s'" % args.save_dir)
            os.makedirs(args.save_dir, exist_ok=True)

            for j, heatmap in enumerate(heatmaps):
                save_path = osp.join(args.save_dir, file_name + "-heatmap-%s.png" % j)
                save_image(save_path, heatmap * 255.0)
        else:
            for j, heatmap in enumerate(heatmaps):
                disp = (image + cm.jet(heatmap)[:, :, 2::-1]) / 2
                cv2.imshow(file_name + "-heatmap-%s" % j, disp)
            cv2.waitKey(0)

    else:
        dataset, target_project = parse_full_revpath(args.target or "project", project)
        if target_project:
            scope_add(target_project)

        log.info("Running inference explanation for '%s'" % args.target)

        for item in dataset:
            image = item.image.data
            if image is None:
                log.warning("Item %s does not have image data. Skipping.", item.id)
                continue

            heatmap_iter = rise.apply(image)

            image = image / 255.0
            heatmaps = next(heatmap_iter)

            if args.save_dir is not None:
                log.info("Saving inference heatmaps to '%s'" % args.save_dir)
                os.makedirs(args.save_dir, exist_ok=True)

                for j, heatmap in enumerate(heatmaps):
                    save_image(
                        osp.join(args.save_dir, item.id + "-heatmap-%s.png" % j),
                        heatmap * 255.0,
                        create_dir=True,
                    )

            if not args.save_dir or args.display:
                for j, heatmap in enumerate(heatmaps):
                    disp = (image + cm.jet(heatmap)[:, :, 2::-1]) / 2
                    cv2.imshow(item.id + "-heatmap-%s" % j, disp)
                cv2.waitKey(0)

    return 0



# Copyright (C) 2021-2022 Intel Corporation
#
# SPDX-License-Identifier: MIT

# The format is described here:
# https://docs.supervise.ly/data-organization/00_ann_format_navi

from __future__ import annotations

import logging as log
import os
import os.path as osp
import shutil
import uuid
from datetime import datetime

from datumaro.components.annotation import AnnotationType, LabelCategories
from datumaro.components.dataset import ItemStatus
from datumaro.components.dataset_base import DatasetItem, IDataset
from datumaro.components.errors import MediaTypeError
from datumaro.components.exporter import Exporter
from datumaro.components.media import PointCloud
from datumaro.util import cast, dump_json_file

from .format import PointCloudPath


class _SuperviselyPointCloudDumper:
    def __init__(self, extractor: IDataset, context: SuperviselyPointCloudExporter):
        self._extractor = extractor
        self._context = context

        timestamp = str(datetime.now())
        self._default_user_info = {
            "labelerLogin": "",
            "createdAt": timestamp,
            "updatedAt": timestamp,
        }

        self._key_id_data = {"tags": {}, "objects": {}, "figures": {}, "videos": {}}

        self._meta_data = {"classes": [], "tags": [], "projectType": "point_clouds"}

        # Meta info contents
        self._tag_meta = {}  # name -> descriptor

        # Registries of item annotations
        self._objects = {}  # id -> key

        self._label_cat = extractor.categories().get(AnnotationType.label, LabelCategories())

    def _write_related_images(self, item):
        img_dir = self._related_images_dir

        for img in item.media.extra_images:
            name = osp.splitext(osp.basename(img.path))[0]
            img_path = osp.join(img_dir, item.id + "_pcd", name + self._find_image_ext(img))
            if img.has_data:
                img.save(img_path)

            img_data = {
                "name": osp.basename(img_path),
                "meta": {
                    "sensorsData": {
                        "extrinsicMatrix": [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
                        "intrinsicMatrix": [0, 0, 0, 0, 0, 0, 0, 0, 0],
                    }
                },
            }

            dump_json_file(osp.join(img_dir, img_path + ".json"), img_data, indent=True)

    def _write_pcd(self, item):
        self._context._save_point_cloud(item, basedir=self._point_cloud_dir)

    def _write_meta(self):
        for tag in self._tag_meta.values():
            if tag["value_type"] is None:
                tag["value_type"] = "any_string"
            tag["classes"] = list(tag["classes"])
        self._meta_data["tags"] = list(self._tag_meta.values())

        dump_json_file(
            osp.join(self._save_dir, PointCloudPath.META_FILE), self._meta_data, indent=True
        )

    def _write_key_id(self):
        objects = self._objects
        key_id_data = self._key_id_data

        key_id_data["objects"] = {v: k for k, v in objects.items()}

        dump_json_file(
            osp.join(self._save_dir, PointCloudPath.KEY_ID_FILE), key_id_data, indent=True
        )

    def _write_item_annotations(self, item):
        key_id_data = self._key_id_data

        item_id = cast(item.attributes.get("frame"), int)
        if item_id is None or self._context._reindex:
            item_id = len(key_id_data["videos"]) + 1

        item_key = str(uuid.uuid4())
        key_id_data["videos"][item_key] = item_id

        item_user_info = {
            k: item.attributes.get(k, default_v) for k, default_v in self._default_user_info.items()
        }

        item_ann_data = {
            "description": item.attributes.get("description", ""),
            "key": item_key,
            "tags": [],
            "objects": [],
            "figures": [],
        }
        self._export_item_attributes(item, item_ann_data, item_user_info)
        self._export_item_annotations(item, item_ann_data, item_user_info)

        ann_path = osp.join(self._ann_dir, item.id + ".pcd.json")
        os.makedirs(osp.dirname(ann_path), exist_ok=True)
        dump_json_file(ann_path, item_ann_data, indent=True)

    def _export_item_attributes(self, item, item_ann_data, item_user_info):
        for attr_name, attr_value in item.attributes.items():
            if attr_name in PointCloudPath.SPECIAL_ATTRS:
                continue

            attr_value = self._encode_attr_value(attr_value)

            tag = self._register_tag(attr_name, value=attr_value, applicable_type="imagesOnly")

            if tag["applicable_type"] != "imagesOnly":
                tag["applicable_type"] = "all"

            value_type = self._define_attr_type(attr_value)
            if tag["value_type"] is None:
                tag["value_type"] = value_type
            elif tag["value_type"] != value_type:
                raise Exception(
                    "Item %s: mismatching "
                    "value types for tag %s: %s vs %s"
                    % (item.id, attr_name, tag["value_type"], value_type)
                )

            tag_key = str(uuid.uuid4())
            item_ann_data["tags"].append(
                {
                    "key": tag_key,
                    "name": attr_name,
                    "value": attr_value,
                    **item_user_info,
                }
            )

            # only item attributes are listed in the key_id file
            # meta tag ids have no relation to key_id tag ids!
            tag_id = len(self._key_id_data["tags"]) + 1
            self._key_id_data["tags"][tag_key] = tag_id

    def _export_item_annotations(self, item, item_ann_data, item_user_info):
        objects = self._objects
        tags = self._tag_meta
        label_cat = self._label_cat
        key_id_data = self._key_id_data

        image_objects = set()
        for ann in item.annotations:
            if not ann.type == AnnotationType.cuboid_3d:
                continue

            obj_id = cast(ann.attributes.get("track_id", ann.id), int)
            if obj_id is None:
                # should not be affected by reindex
                # because it is used to match figures,
                # including different frames
                obj_id = len(self._objects) + 1

            object_key = objects.setdefault(obj_id, str(uuid.uuid4()))
            object_label = label_cat[ann.label].name
            if obj_id not in image_objects:
                ann_user_info = {
                    k: ann.attributes.get(k, default_v) for k, default_v in item_user_info.items()
                }

                obj_ann_data = {
                    "key": object_key,
                    "classTitle": object_label,
                    "tags": [],
                    "objects": [],
                    "figures": [],
                    **ann_user_info,
                }

                for attr_name, attr_value in ann.attributes.items():
                    if attr_name in PointCloudPath.SPECIAL_ATTRS:
                        continue

                    attr_value = self._encode_attr_value(attr_value)

                    tag = tags.get(attr_name)
                    if tag is None:
                        if self._context._allow_undeclared_attrs:
                            tag = self._register_tag(attr_name, applicable_type="objectsOnly")
                            tags[attr_name] = tag
                        else:
                            log.warning(
                                "Item %s: skipping undeclared "
                                "attribute '%s' for label '%s' "
                                "(allow with --allow-undeclared-attrs option)",
                                item.id,
                                attr_name,
                                object_label,
                            )
                            continue

                    if tag["applicable_type"] == "imagesOnly":
                        tag["applicable_type"] = "all"
                    elif tag["applicable_type"] == "objectsOnly" and tag["classes"]:
                        tag["classes"].add(object_label)

                    value_type = self._define_attr_type(attr_value)
                    if tag["value_type"] is None:
                        tag["value_type"] = value_type
                    elif tag["value_type"] != value_type:
                        raise Exception(
                            "Item %s: mismatching "
                            "value types for tag %s: %s vs %s"
                            % (item.id, attr_name, tag["value_type"], value_type)
                        )

                    tag_key = str(uuid.uuid4())
                    obj_ann_data["tags"].append(
                        {
                            "key": tag_key,
                            "name": attr_name,
                            "value": attr_value,
                            **ann_user_info,
                        }
                    )

                item_ann_data["objects"].append(obj_ann_data)

                image_objects.add(obj_id)

            figure_key = str(uuid.uuid4())
            item_ann_data["figures"].append(
                {
                    "key": figure_key,
                    "objectKey": object_key,
                    "geometryType": "cuboid_3d",
                    "geometry": {
                        "position": {
                            "x": float(ann.position[0]),
                            "y": float(ann.position[1]),
                            "z": float(ann.position[2]),
                        },
                        "rotation": {
                            "x": float(ann.rotation[0]),
                            "y": float(ann.rotation[1]),
                            "z": float(ann.rotation[2]),
                        },
                        "dimensions": {
                            "x": float(ann.scale[0]),
                            "y": float(ann.scale[1]),
                            "z": float(ann.scale[2]),
                        },
                    },
                    **ann_user_info,
                }
            )
            figure_id = ann.id
            if self._context._reindex or figure_id is None:
                figure_id = len(key_id_data["figures"]) + 1
            key_id_data["figures"][figure_key] = figure_id

    @staticmethod
    def _encode_attr_value(v):
        if v is True or v is False:  # use is to check the type too
            v = str(v).lower()
        return v

    @staticmethod
    def _define_attr_type(v):
        if isinstance(v, (int, float)):
            t = "any_number"
        else:
            t = "any_string"
        return t

    def _register_tag(self, name, **kwargs):
        tag = {
            "name": name,
            "value_type": None,
            "color": "",
            "id": len(self._tag_meta) + 1,
            "hotkey": "",
            "applicable_type": "all",
            "classes": set(),
        }
        tag.update(kwargs)
        return self._tag_meta.setdefault(name, tag)

    def _make_dirs(self):
        save_dir = self._context._save_dir
        os.makedirs(save_dir, exist_ok=True)
        self._save_dir = save_dir

        base_dir = osp.join(self._save_dir, PointCloudPath.BASE_DIR)
        os.makedirs(base_dir, exist_ok=True)

        ann_dir = osp.join(base_dir, PointCloudPath.ANNNOTATION_DIR)
        os.makedirs(ann_dir, exist_ok=True)
        self._ann_dir = ann_dir

        point_cloud_dir = osp.join(base_dir, PointCloudPath.POINT_CLOUD_DIR)
        os.makedirs(point_cloud_dir, exist_ok=True)
        self._point_cloud_dir = point_cloud_dir

        related_images_dir = osp.join(base_dir, PointCloudPath.RELATED_IMAGES_DIR)
        os.makedirs(related_images_dir, exist_ok=True)
        self._related_images_dir = related_images_dir

    def _init_meta(self):
        for attr in self._label_cat.attributes:
            self._register_tag(attr, applicable_type="objectsOnly")

        for idx, label in enumerate(self._label_cat):
            self._meta_data["classes"].append(
                {
                    "id": idx,
                    "title": label.name,
                    "color": "",
                    "shape": "cuboid_3d",
                    "geometry_config": {},
                }
            )

            for attr in label.attributes:
                tag = self._register_tag(attr, applicable_type="objectsOnly")
                tag["classes"].add(label.name)

    def _find_image_ext(self, image):
        src_ext = image.ext
        return self._context._image_ext or src_ext or self._context._default_image_ext

    def dump(self):
        self._make_dirs()

        self._init_meta()

        for item in self._context._extractor:
            if self._context._save_media:
                if item.media:
                    self._write_pcd(item)
                else:
                    log.debug("Item '%s' has no point cloud info", item.id)

                if item.media and item.media.extra_images:
                    self._write_related_images(item)
                else:
                    log.debug("Item '%s' has no related images info", item.id)

            self._write_item_annotations(item)

        self._write_meta()
        self._write_key_id()


class SuperviselyPointCloudExporter(Exporter):
    NAME = "sly_pointcloud"
    DEFAULT_IMAGE_EXT = PointCloudPath.DEFAULT_IMAGE_EXT

    @classmethod
    def build_cmdline_parser(cls, **kwargs):
        parser = super().build_cmdline_parser(**kwargs)
        parser.add_argument(
            "--reindex",
            action="store_true",
            help="Assign new indices to frames (default: %(default)s)",
        )
        parser.add_argument(
            "--allow-undeclared-attrs",
            action="store_true",
            help="Write annotation attributes even if they are not present in "
            "the input dataset metainfo (default: %(default)s)",
        )
        return parser

    def __init__(self, extractor, save_dir, reindex=False, allow_undeclared_attrs=False, **kwargs):
        super().__init__(extractor, save_dir, **kwargs)

        self._reindex = reindex
        self._allow_undeclared_attrs = allow_undeclared_attrs

    def apply(self):
        if self._extractor.media_type() and self._extractor.media_type() is not PointCloud:
            raise MediaTypeError("Media type is not an image")

        if 1 < len(self._extractor.subsets()):
            log.warning(
                "Supervisely pointcloud format supports only a single "
                "subset. Subset information will be ignored on export."
            )

        _SuperviselyPointCloudDumper(self._extractor, self).dump()

    @classmethod
    def patch(cls, dataset, patch, save_dir, **kwargs):
        conv = cls(patch.as_dataset(dataset), save_dir=save_dir, **kwargs)
        conv.apply()

        for (item_id, subset), status in patch.updated_items.items():
            if status != ItemStatus.removed:
                item = patch.data.get(item_id, subset)
            else:
                item = DatasetItem(item_id, subset=subset)

            if not (status == ItemStatus.removed or not item.media):
                continue

            pcd_name = conv._make_pcd_filename(item)

            ann_path = osp.join(
                save_dir,
                PointCloudPath.BASE_DIR,
                PointCloudPath.ANNNOTATION_DIR,
                pcd_name + ".json",
            )
            if osp.isfile(ann_path):
                os.remove(ann_path)

            pcd_path = osp.join(
                save_dir, PointCloudPath.BASE_DIR, PointCloudPath.POINT_CLOUD_DIR, pcd_name
            )
            if osp.isfile(pcd_path):
                os.remove(pcd_path)

            images_dir = osp.join(
                save_dir,
                PointCloudPath.BASE_DIR,
                PointCloudPath.RELATED_IMAGES_DIR,
                osp.splitext(pcd_name)[0] + "_pcd",
            )
            if osp.isdir(images_dir):
                shutil.rmtree(images_dir)



# Copyright (C) 2019-2020 Intel Corporation
#
# SPDX-License-Identifier: MIT


class DetectionApiPath:
    IMAGES_DIR = "images"
    ANNOTATIONS_DIR = "annotations"

    DEFAULT_IMAGE_EXT = ".jpg"
    IMAGE_EXT_FORMAT = {".jpg": "jpeg", ".jpeg": "jpeg", ".png": "png"}

    LABELMAP_FILE = "label_map.pbtxt"



# Copyright (C) 2020-2021 Intel Corporation
#
# SPDX-License-Identifier: MIT

import argparse
import logging as log

from datumaro.components.environment import Environment
from datumaro.components.errors import ProjectNotFoundError
from datumaro.components.validator import TaskType
from datumaro.util import dump_json_file
from datumaro.util.scope import scope_add, scoped

from ..util import MultilineFormatter
from ..util.errors import CliException
from ..util.project import generate_next_file_name, load_project, parse_full_revpath


def build_parser(parser_ctor=argparse.ArgumentParser):
    parser = parser_ctor(
        help="Validate project",
        description="""
        Validates a dataset according to the task type and
        reports summary in a JSON file.|n
        Target dataset is specified by a revpath. The full syntax is:|n
        - Dataset paths:|n
        |s|s- <dataset path>[ :<format> ]|n
        - Revision paths:|n
        |s|s- <project path> [ @<rev> ] [ :<target> ]|n
        |s|s- <rev> [ :<target> ]|n
        |s|s- <target>|n
        |n
        Both forms use the -p/--project as a context for plugins. It can be
        useful for dataset paths in targets. When not specified, the current
        project's working tree is used.|n
        |n
        Examples:|n
        - Validate a project's subset as a classification dataset:|n |n
        |s|s%(prog)s -t classification -s train
        """,
        formatter_class=MultilineFormatter,
    )

    task_types = ", ".join(t.name for t in TaskType)

    def _parse_task_type(s):
        try:
            return TaskType[s.lower()].name
        except:
            raise argparse.ArgumentTypeError(
                "Unknown task type %s. Expected " "one of: %s" % (s, task_types)
            )

    parser.add_argument(
        "_positionals", nargs=argparse.REMAINDER, help=argparse.SUPPRESS
    )  # workaround for -- eaten by positionals
    parser.add_argument(
        "target", default="project", nargs="?", help="Target dataset revpath (default: project)"
    )
    parser.add_argument(
        "-t",
        "--task",
        type=_parse_task_type,
        required=True,
        help="Task type for validation, one of %s" % task_types,
    )
    parser.add_argument(
        "-s", "--subset", dest="subset_name", help="Subset to validate (default: whole dataset)"
    )
    parser.add_argument(
        "-p",
        "--project",
        dest="project_dir",
        help="Directory of the project to validate (default: current dir)",
    )
    parser.add_argument(
        "extra_args",
        nargs=argparse.REMAINDER,
        help="Optional arguments for validator (pass '-- -h' for help)",
    )
    parser.set_defaults(command=validate_command)

    return parser


def get_sensitive_args():
    return {
        validate_command: ["target", "project_dir", "subset_name", "extra_args"],
    }


@scoped
def validate_command(args):
    has_sep = "--" in args._positionals
    if has_sep:
        pos = args._positionals.index("--")
        if 1 < pos:
            raise argparse.ArgumentError(None, message="Expected no more than 1 target argument")
    else:
        pos = 1
    args.target = (args._positionals[:pos] or ["project"])[0]
    args.extra_args = args._positionals[pos + has_sep :]

    show_plugin_help = "-h" in args.extra_args or "--help" in args.extra_args

    project = None
    try:
        project = scope_add(load_project(args.project_dir))
    except ProjectNotFoundError:
        if not show_plugin_help and args.project_dir:
            raise

    if project is not None:
        env = project.env
    else:
        env = Environment()

    try:
        validator_type = env.validators[args.task]
    except KeyError:
        raise CliException("Validator type '%s' is not found" % args.task)

    extra_args = validator_type.parse_cmdline(args.extra_args)

    dataset, target_project = parse_full_revpath(args.target, project)
    if target_project:
        scope_add(target_project)

    dst_file_name = f"validation-report"
    if args.subset_name is not None:
        dataset = dataset.get_subset(args.subset_name)
        dst_file_name += f"-{args.subset_name}"

    validator = validator_type(**extra_args)
    report = validator.validate(dataset)

    def _make_serializable(d):
        for key, val in list(d.items()):
            # tuple key to str
            if isinstance(key, tuple) or isinstance(key, int):
                d[str(key)] = val
                d.pop(key)
            if isinstance(val, dict):
                _make_serializable(val)

    _make_serializable(report)

    dst_file = generate_next_file_name(dst_file_name, ext=".json")
    log.info("Writing project validation results to '%s'" % dst_file)
    dump_json_file(dst_file, report, indent=True, allow_numpy=True)
